#!/usr/bin/env python3
import numpy as np
import numba as nb
import heapq
import packaging.version

# Note: Reflected lists are deprecated in numba >=0.51. Typed lists are available from numba >=0.45.
# The heapq used for the priority queue in the Dijkstra algorithm does not accept typed list though (numba issue #4926).
# Thus, the Numba version needs to be pinned. Numba <=0.50 issues scheduled deprecation warnings
# but does run the Dijkstra implementation in nopython mode with Python 3.7 and 3.8 (not 3.9, numba issue #6345).
# Numba 0.50 requires numpy<1.20 (numba issue #6041)

_hasTypedList = packaging.version.parse(nb.__version__) >= packaging.version.parse('0.45.0')

_dist_list = np.sqrt(np.array([1, 2, 3, 5, 6]))  # reduction to 74 neighbors


class Grid3D:
    """Class object holding a 3D grid

    Parameters
    ----------
    mol_xyzr : ndarray
               array of x-,y-,z-coordinates and VdW radii with a shape [n_atoms, 4]
    attach_xyz : ndarray
                one-dimensional array of x-,y-,z-coordinates of the attachment point
                (corresponds to the center of the grid)
    linker_length : float
                    length of the dye linker in Angstrom
    linker_width : float
                   diameter of the dye linker in Angstrom
    dye_radii : ndarray([3,1])
                array of dye radii in Angstrom with shape [3,1]
    grid_spacing : float
        spacing between grid points (in A)

    Notes
    -----
    Attributes of the LabelLib class
        - discStep : float  (the grid spacing)
        - originXYZ : numpy.array  (x-/y-/z-coordinate of the grid origin)
        - shape : numpy.array      (number of grid points in x-/y-/z-direction)
        - grid : numpy.array       (flattened list of grid point values)
    """

    def __init__(self, mol_xyzr, attach_xyz, linker_length, linker_width, dye_radii, grid_spacing, simulation_type):
        self.discStep = grid_spacing
        self.attach_xyz = attach_xyz
        self.grid_3d, self.originXYZ, self.originAdj = self.make_grid(attach_xyz, linker_length, grid_spacing)
        self.shape = np.array(self.grid_3d.shape)
        self.halfCubeLength = min(self.shape) * self.discStep * 0.5
        self.grid_3d = self.block_molecule(mol_xyzr, 0.5 * linker_width)
        maxR = 3 * grid_spacing
        maxR_source = linker_width + grid_spacing
        self.grid_3d = self.dijkstra_init(maxR, maxR_source)
        self.grid_3d = self.setAboveTreshold(self.grid_3d, linker_length, -4)
        self.grid_3d = self.setAboveTreshold(self.grid_3d, 0, 1)
        if simulation_type == 'AV1':
            dye_radii = [dye_radii[0]]
        self.grid_3d = self.excludeConcentricSpheres(mol_xyzr, dye_radii, 1)
        self.grid = self.grid_3d.flatten(order='F')

    @staticmethod
    @nb.jit(forceobj=True)
    def make_grid(attach_xyz, linker_length, grid_spacing):
        """
        Build a 3D grid around the attachment point

        Parameters
        ----------
        attach_xyz : ndarray
                     one-dimensional array of x-,y-,z-coordinates of the attachment point
                     (corresponds to the center of the grid)
        linker_length : float
                        length of the dye linker in Angstrom
        grid_spacing : float
            spacing between grid points (in A)

        Returns
        -------
        grid_3d : ndarray
                  3-dimensional array of grid points with a shape of 2*ll_padRound+1
        xyz_min : ndarray
        """
        ll_pad = linker_length + 3 * grid_spacing  # pad the linker_length
        ll_padRound = np.ceil(ll_pad / grid_spacing) * grid_spacing  # round to next higher grid value
        xyz_min = attach_xyz - ll_padRound
        originAdj = xyz_min - 0.5 * grid_spacing
        gridptsPerEdge = 2 * int(ll_padRound / grid_spacing + 0.5)  # + 1
        grid_3d = np.full([gridptsPerEdge] * 3, np.inf)
        for i in range(gridptsPerEdge):
            for j in range(gridptsPerEdge):
                for k in range(gridptsPerEdge):
                    ijk = np.array([i, j, k])
                    maxRSq = int(linker_length**2 / grid_spacing**2 + 0.5)
                    ijk0 = (attach_xyz - xyz_min) / grid_spacing
                    dSq = np.sum((ijk - ijk0)**2)
                    if dSq > maxRSq:
                        grid_3d[(i, j, k)] = -1
        return grid_3d, xyz_min, originAdj

    @staticmethod
    def _xyz2idx(xyz, originAdj, grid_spacing, decimals=6):
        """Get the ijk grid indices for a set of xyz values

        Parameters
        ----------
        xyz :
        originAdj :
        grid_spacing : float
            spacing between grid points (in A)

        Note
        ----
        Round to n-decimal places (here: 4) before casting to integer
        """
        return np.round(((xyz - originAdj) / grid_spacing), decimals).astype(np.int64)

    def block_molecule(self, mol_xyzr, extraClash):
        """Block the grid points which are within the VdW radius of any atom of the biomolecule plus an extra clash radius

        Parameters
        ----------
        mol_xyzr : ndarray
            array of x-,y-,z-coordinates and VdW radii with a shape [n_atoms, 4]
        extraClash : float
            clash radius added to the VdW radius

        Returns
        -------
        ndarray
            3-dimensional array of grid points with a shape of 2*adjL+1
        """
        maxVdW_extraClash = np.max(mol_xyzr[:, 3]) + extraClash
        if _hasTypedList:
            neighbor_list = nb.typed.List()
            [neighbor_list.append(n) for n in self.sortedNeighborIdx(maxVdW_extraClash)]
        else:
            neighbor_list = self.sortedNeighborIdx(maxVdW_extraClash)
        ijk_atom = self._xyz2idx(mol_xyzr[:, 0:3], self.originAdj, self.discStep)
        outDistSq = (self.halfCubeLength + maxVdW_extraClash)**2
        distSq = np.sum((mol_xyzr[:, 0:3] - self.attach_xyz)**2, 1)
        grid_3d = self._carve_VdWextraClash(self.grid_3d, mol_xyzr, neighbor_list, ijk_atom, extraClash, distSq,
                                            outDistSq, self.shape)
        return grid_3d

    @staticmethod
    @nb.jit(nopython=True)
    def _carve_VdWextraClash(grid_3d, mol_xyzr, neighbor_list, ijk_atom, extraClash, distSq, outDistSq, grid_shape):
        """
        Loop through the atoms and assign -1 to all grid values that are within the atoms VdW radius
        plus an extraClash value
        """
        for m in range(mol_xyzr.shape[0]):
            if distSq[m] > outDistSq:
                continue
            for n in neighbor_list:
                if n[0] > mol_xyzr[m, 3] + extraClash:
                    break
                ijk_n = ijk_atom[m] + n[1]
                i, j, k = ijk_n
                if not np.any(ijk_n < 0) and not np.any(grid_shape - ijk_n <= 0):
                    grid_3d[i, j, k] = -1
        return grid_3d

    def dijkstra_init(self, maxR, maxR_source):
        """
        Initialize the Dijkstra search algorithm

        Parameters
        ----------
        maxR : float
               radius within to search for neighbors
        maxRsource : float
                     radius within to search for neighbors in the first round of the Dijkstra algorithm

        Returns
        -------
        grid_3d : ndarray
                  3-dimensional array of grid points with a shape of 2*adjL+1
        """
        ai, aj, ak = self._xyz2idx(self.attach_xyz, self.originAdj, self.discStep)
        self.grid_3d[ai, aj, ak] = 0
        priority_queue = [(0.0, (ai, aj, ak))]
        if _hasTypedList:
            # priority_queue = nb.typed.List()
            # [priority_queue.append(x) for x in [(0.0, (ai, aj, ak))]]
            edges_ess = nb.typed.List()
            [edges_ess.append(x) for x in self.essential_neighbors(self.sortedNeighborIdx(maxR), _dist_list)]
            edges_src = nb.typed.List()
            [edges_src.append(x) for x in self.sortedNeighborIdx(maxR_source)]
        else:
            edges_ess = self.essential_neighbors(self.sortedNeighborIdx(maxR), _dist_list)
            edges_src = self.sortedNeighborIdx(maxR_source)
        grid_3d = self.dijkstra(self.grid_3d, edges_ess, edges_src, priority_queue, (ai, aj, ak))
        return grid_3d

    def sortedNeighborIdx(self, maxR):
        """
        Build a neighbor list with a maximal extent given by maxR and sorted by increasing distance
        from the origin

        Parameters
        ----------
        maxR : float
               radius within to search for neighbors

        Returns
        -------
        idxs : list of 2-tuples of numpy.ndarray and float
               the tuples contain the ijk indices and the distance from the origin (0,0,0)
        """
        idxs = self._neighborIdx(maxR, self.discStep)
        idxs.sort(key=lambda x: x[0])
        return idxs

    @staticmethod
    @nb.jit(forceobj=True)
    def _neighborIdx(maxR, grid_spacing):
        """Build a list of neighboring indices to the origin (0,0,0)

        Parameters
        ----------
        maxR : float
            radius within which to search for neighbors
        grid_spacing : float
            spacing between grid points (in A)

        Returns
        -------
        idxs : list of 2-tuples of numpy.ndarray and float
            the tuples contain the ijk indices and the distance from the origin (0,0,0), e.g. [(1.0,array([-1,0,0]))]

        Notes
        -----
        The neighbor list is built from an origin set at (0,0,0). imaxR defines the number of indices
        along one axis (i,j or k) to build the index cube.
        For imaxR=3, first a cube of 7*7*7=343 indices is built (1).
        This is reduced in a second step to 123 indices that have a distance < imaxR from the origin (2).
        Finally, they are reduced 74 essential neighbors by the distances in _dist_list (3).
        (this last reduction speeds up the Dikstra algorithm due to the shorter neighbor list without loosing much
        accuracy)
        """
        idxs = []
        maxRSq = maxR**2
        imaxR = int(maxR / grid_spacing + 0.5)  # rounds to next integer
        # (1) build a cube with (2*imaxR+1)**3 indices
        for k in range(-imaxR, imaxR + 1, 1):
            for j in range(-imaxR, imaxR + 1, 1):
                for i in range(-imaxR, imaxR + 1, 1):
                    ijk = np.array([i, j, k], dtype=int)
                    # (2) only accept indices with 0 < dS < imaxRSq
                    dSq = np.sum(ijk**2) * grid_spacing**2
                    if dSq <= maxRSq:  # and dSq > 0:
                        d = np.sqrt(dSq)
                        idxs.append((d, ijk))
        return idxs

    def essential_neighbors(self, idxs, dist_list):
        """
        Reduce the neighbor list to those indices with a distance from the origin featured in dist_list

        Parameters
        ----------
        idxs : list of 2-tuples of numpy.ndarray and float
               the tuples contain the ijk indices and the distance from the origin (0,0,0)

        idxs_ess : list of 2-tuples of numpy.ndarray and float
                   list n essential neighbors from the origin (0,0,0)

        Notes
        -----
        If dist_list=sqrt([1, 2, 3, 5, 6]) then the the neighbor list is reduced to 74 neighbors
        which make up a spherical shape and are a good compromise between neighbor space and time efficiency
        in the Dijkstra algorithm. The smaller this list the faster the Dijkstra algorithm will run
        at the expense of the accuracy of the resulting accessible volume (i.e. its "sphericalness")
        """
        idxs_ess = []
        for e in idxs:
            minDiff = min(abs(e[0] - dist_list * self.discStep))
            if minDiff < (0.01 * self.discStep):
                idxs_ess.append(e)
        return idxs_ess

    @staticmethod
    @nb.jit(nopython=True)
    def dijkstra(grid_3d, edges_ess, edges_src, priority_queue, start_ijk):
        """
        Djikstra algorithm with a priority queue

        Parameters
        ----------
        grid_3d : numpy.ndarray
                  3-dimensional array of grid points with a shape given by n_xyz
        edges_ess : nb.typed.List of 2-tuples of numpy.ndarray and float
                    list n essential neighbors from the origin (0,0,0)
        edges_src : list of 2-tuples of numpy.ndarray and float
                    nb.typed.List of neighbors in the initialization round of the Dijkstra algorithm
        priority_queue : list of 2-tuples of numpy.ndarray and float
                         list with distance and index of origin

        Returns
        -------
        grid_3d : numpy.ndarray([nx,ny,nz])
                  3-dimensional array of grid points with a shape given by n_xyz
        """
        while priority_queue:
            r, idx = heapq.heappop(priority_queue)
            if r > grid_3d[idx]:
                continue
            if idx == start_ijk:
                edges = edges_src
            else:
                edges = edges_ess
            for e in edges:
                i = idx[0] + e[1][0]
                j = idx[1] + e[1][1]
                k = idx[2] + e[1][2]
                val = grid_3d[i, j, k]
                if val < 0:
                    continue
                t_r = r + e[0]
                if t_r < val:
                    grid_3d[i, j, k] = t_r
                    heapq.heappush(priority_queue, (t_r, (i, j, k)))
        return grid_3d

    @staticmethod
    @nb.jit(nopython=True)
    def setAboveTreshold(grid_3d, treshold, new_value):
        """
        Reassign grid values which are above a specified treshold

        Parameters
        ----------
        treshold : float
        new_value : float

        Returns
        -------
        grid_3d : numpy.ndarray([nx,ny,nz])
                  3-dimensional array of grid points with a shape given by n_xyz
        """
        for i in range(grid_3d.shape[0]):
            for j in range(grid_3d.shape[1]):
                for k in range(grid_3d.shape[2]):
                    if grid_3d[(i, j, k)] > treshold:
                        grid_3d[(i, j, k)] = new_value
        return grid_3d

    def excludeConcentricSpheres(self, mol_xyzr, dye_radii, maxRho):
        """
        Parameters
        ----------
        mol_xyzr : numpy.ndarray
                   array of x-,y-,z-coordinates and VdW radii with a shape [n_atoms, 4]
        dye_radii : ndarray([3,1])
                    array of dye radii in Angstrom with shape [3,1]
        maxRho : float
                 maximum grid value in the free volume

        Returns
        -------
        grid_3d : numpy.ndarray([nx,ny,nz])
                  3-dimensional array of grid points with a shape given by n_xyz
        """
        maxVdW_extraClash = np.max(mol_xyzr[:, 3]) + np.max(dye_radii)
        if _hasTypedList:
            neighbor_list = nb.typed.List()
            [neighbor_list.append(n) for n in self.sortedNeighborIdx(maxVdW_extraClash)]
            dye_radii_sorted = nb.typed.List()
            [dye_radii_sorted.append(x) for x in sorted(dye_radii)]
        else:
            neighbor_list = self.sortedNeighborIdx(maxVdW_extraClash)
            dye_radii_sorted = sorted(dye_radii)
        rhos = np.linspace(0, maxRho, len(dye_radii) + 1)
        ijk_atom = self._xyz2idx(mol_xyzr[:, 0:3], self.originAdj, self.discStep)
        outdistSq = (self.halfCubeLength + maxVdW_extraClash)**2
        distSq = np.sum((mol_xyzr[:, 0:3] - self.attach_xyz)**2, 1)
        grid_3d = self._assignRho(self.grid_3d, mol_xyzr, neighbor_list, ijk_atom, dye_radii_sorted, rhos, distSq,
                                  outdistSq, self.shape)
        return grid_3d

    @staticmethod
    @nb.jit(nopython=True)
    def _assignRho(grid_3d, mol_xyzr, neighbor_list, ijk_atom, dye_radii_sorted, rhos, distSq, outdistSq, grid_shape):
        """
        Loop through the atoms and radii and reassign the grid values with a number 0 < rho < 1
        """
        n = len(neighbor_list)
        n_radii = len(dye_radii_sorted)
        for m in range(mol_xyzr.shape[0]):
            if distSq[m] > outdistSq:
                continue
            else:
                s = 0
                for r in range(n_radii):
                    effR = dye_radii_sorted[r] + mol_xyzr[m, 3]
                    while neighbor_list[s][0] <= effR:
                        ijk_n = ijk_atom[m] + neighbor_list[s][1]
                        i, j, k = ijk_n
                        if not np.any(ijk_n < 0) and not np.any(grid_shape - ijk_n <= 0):
                            gridval = grid_3d[i, j, k]
                            grid_3d[i, j, k] = min(rhos[r], gridval)
                        s += 1
                        if s >= n:
                            break
        return grid_3d

    def addWeights(self, das_xyzrm):
        """
        Assign a weight to the grid values which are part of the contact volume

        Parameters
        ----------
        das_xyzrm : numpy.ndarray([5,n_atoms])
                    array of marked coordinates and padded vdW radii (n_atoms = number of atoms in mdtraj.Trajectory)

        Returns
        -------
        grid_3d : numpy.ndarray([nx,ny,nz])
                  3-dimensional array of grid points with a shape given by n_xyz
        """
        maxClash = np.max(das_xyzrm[:, 3]) + self.discStep
        if _hasTypedList:
            neighbor_list = nb.typed.List()
            [neighbor_list.append(n) for n in self.sortedNeighborIdx(maxClash)]
        else:
            neighbor_list = self.sortedNeighborIdx(maxClash)
        ijk_atom = self._xyz2idx(das_xyzrm[:, 0:3], self.originAdj, self.discStep)
        outDistSq = (self.halfCubeLength + maxClash)**2
        distSq = np.sum((das_xyzrm[:, 0:3] - self.attach_xyz)**2, 1)
        grid_3d = self._assignDensity(self.grid_3d, das_xyzrm, neighbor_list, ijk_atom, distSq, outDistSq, self.shape)
        return grid_3d

    @staticmethod
    @nb.jit(nopython=True)
    def _assignDensity(grid_3d, das_xyzrm, neighbor_list, ijk_atom, distSq, outDistSq, grid_shape):
        """
        Loop through the atoms and assign the density
        """
        n = len(neighbor_list)
        for m in range(das_xyzrm.shape[0]):
            if distSq[m] > outDistSq:
                continue
            else:
                s = 0
                while neighbor_list[s][0] <= das_xyzrm[m, 3]:
                    ijk_n = ijk_atom[m] + neighbor_list[s][1]
                    i, j, k = ijk_n
                    if not np.any(ijk_n < 0) and not np.any(grid_shape - ijk_n <= 0):
                        gridval = grid_3d[i, j, k]
                        if gridval > 0:
                            grid_3d[i, j, k] += das_xyzrm[m, 4]
                    s += 1
                    if s >= n:
                        break
        return grid_3d



#!/usr/bin/env python3

import numpy as np
import mdtraj as md
import copy

try:
    from fretraj import export
except ImportError: # for PyMOL plugin
    from . import export

class Plumed:

    def __init__(self, structure, volumes, selection='name P', cutoff=1, pseudo_element=md.element.bromine):
        self.struct = copy.deepcopy(structure)
        chain = self.struct.top.add_chain()
        pseudo_serials = []
        self.dist_restraints = []
        self.xyz = self.struct.xyz * 10
        for volume in volumes:
            res = self.struct.top.add_residue(pseudo_element.symbol, chain, max([res.resSeq for res in self.struct.top.residues])+1)
            atom = self.struct.top.add_atom(pseudo_element.symbol, pseudo_element, res, self.struct.top.n_atoms+1)
            pseudo_serials.append(atom.serial)
            self.dist_restraints.append(self.distance_restraint(atom.serial, volume, selection, cutoff))
            self.xyz = np.hstack((self.xyz, np.array([[volume.acv.mp]])))
        self.n_atoms = self.struct.top.n_atoms


    def distance_restraint(self, pseudo_serial, volume, selection='name P', cutoff=15):
        """
        cutoff : float
                 radius within which to search for atoms in the selection (in Angstrom)
        """
        attach_id_mdtraj = volume.attach_id-1
        neighbor_idx = md.compute_neighbors(self.struct, cutoff / 10, [attach_id_mdtraj])[0]
        atomsele_idx = volume.structure.top.select(selection)
        atoms_idx = np.array([idx for idx in neighbor_idx if idx in atomsele_idx])
        atoms_ser = atoms_idx+1
        atoms_xyz = volume.structure.xyz[0,atoms_idx,:] * 10
        distances = np.sqrt(np.sum((volume.acv.mp - atoms_xyz)**2, axis=1)) / 10
        return {'pseudo_atom': {'xyz': volume.acv.mp / 10}, 
                'attach_atoms': {'serial': atoms_ser, 'xyz': volume.structure.xyz[0,atoms_idx,:]},
                'distances': distances}

    def write_plumed(self, filename, R_mp, kappa_dist=1000, Rmp_kappa=500):
        """
        """
        plumed_str = ''
        k = 1
        plumed_str += 'mp1: GROUP ATOMS=xxx\n'
        plumed_str += 'mp2: GROUP ATOMS=yyy\n\n'
        for dr in self.dist_restraints:
            for i,attach_atom in enumerate(dr['attach_atoms']['serial']):
                plumed_str += 'd{:d}{:d}: DISTANCE ATOMS=mp{:d},{}\n'.format(k,i,k,attach_atom)
            for i,dist in enumerate(dr['distances']):
                plumed_str += 'r{:d}{:d}: RESTRAINT ARG=d{:d}{:d} AT={:0.1f} KAPPA={}\n'.format(k,i,k,i,dist,kappa_dist)
            plumed_str += '\n'
            k += 1
        plumed_str += 'WHOLEMOLECULES ENTITY0=1-{:d}\n'.format(self.n_atoms)
        plumed_str += 'Rmp: DISTANCE ATOMS=mp1,mp2 NOPBC\n'
        plumed_str += 'RESTRAINT ARG=Rmp AT={:0.1f} KAPPA={}\n'.format(R_mp, Rmp_kappa)
        with open(filename, 'w') as f:
            f.write(plumed_str)

    def write_pseudo(self, pdb_filename, itp_filename, pseudo_element='MP'):
        with open(pdb_filename, 'w') as f:
            pdbstr = export._pdb_format.format('ATOM', 1, pseudo_element, ' ', pseudo_element, ' ', 1, ' ', 0.000, 0.000, 0.000, 1, 0, ' ', ' ')
            f.write(pdbstr)
        with open(itp_filename, 'w') as f:
            itpstr = '''[ moleculetype ]
; molname       nrexcl
MP             1       ; pseudoatom representing the dye mean position

[ atoms ]
; id    at type         res nr  residue name     at name  cg nr  charge
1       MP              1       MP             MP       1      0.00000
'''
            f.write(itpstr)
        print('(1) Copy the \"{}\" file into your force field directory\n'.format(itp_filename.split('/')[-1]))
        print('(2) Add the following lines to your \"topology.top\" file:\n; Include topology for dye mean position\n#include "<path/to/forcefield>/{}"\n\n#ifdef POSRES\n[ position_restraints ]\n1     1  1000  1000  1000\n#endif\n'.format(itp_filename.split('/')[-1]))
        print('(3) Add the following line to your \"atomtypes.atp\" file in the force field directory:\n{}        0.00000 ;\n'.format(pseudo_element))
        print('(4) Add the following lines to your \"ffnonbonded.itp\" file in the force field directory:\n; Dummy mass for dye mean position\n{}          35      79.90    0.0000  A   4.64693e-01  2.45414e-01'.format(pseudo_element))


    # def write_pseudoGRO(self, filename):
    #     pdbfile = md.formats.GroTrajectoryFile(filename, mode='w')
    #     pdbfile.write(self.xyz / 10, self.struct.top)


    def write_vmd(self, filename='vis.vmd'):
        """
        Generate a visualization file for VMD containing Tcl API commands
        """
        vmd_vis = '''color change rgb blue    0.20  0.33  0.72
color change rgb red     0.76  0.33  0.29
color change rgb gray    0.71  0.74  0.77
color change rgb orange  0.89  0.50  0.32
color change rgb green   0.42  0.70  0.51
color change rgb violet  0.55  0.39  0.70
axes location off

mol modselect 0 0 all not solvent
mol modstyle 0 0 Licorice 0.300000 12.000000 12.000000
mol modcolor 0 0 ColorID 2
mol modmaterial 0 0 AOChalky

mol addrep 0
mol modselect 1 0 {resname MP}
mol modcolor 1 0 ColorID 0
mol modstyle 1 0 VDW 1.500000 12.000000

set sel [atomselect 0 "resname MP"]
'''
        for i in range(2):
            vmd_vis += 'set pseudoid{:d} [lindex [$sel get index] {:d}]\n'.format(i,i)
            for serial in self.dist_restraints[i]['attach_atoms']['serial']:
                vmd_vis += 'label add Bonds 0/$pseudoid{:d} 0/{:d}\n'.format(i,serial-1)
        vmd_vis += 'label add Bonds 0/$pseudoid0 0/$pseudoid1\n'
        vmd_vis += 'label textsize 0.5\n'
        with open(filename, 'w') as f:
            f.write(vmd_vis)


    def write_pymol(self, filename='vis.py'):
        """
        Generate a visualization file for PyMOL containing Python API commands
        """
        pymol_vis = '''cmd.hide(\'lines\')
cmd.show(\'sticks\', \'all and not hydrogen\')
cmd.remove(\'solvent or (inorganic and not resn MP)\')
cmd.set(\'cartoon_ring_mode\',3)
cmd.set(\'cartoon_ring_finder\',0)
cmd.color(\'gray80\')
cmd.show(\'spheres\', \'resn MP\')
cmd.set(\'sphere_scale\', 1.5, \'resn MP\')
cmd.color('skyblue', \'resn MP\')
cmd.set(\'dash_radius\', 0.3)
cmd.set(\'dash_gap\', 1)
'''
        for i in range(2):
            pymol_vis +='cmd.pseudoatom(\'p{:d}\', pos=[{:0.2f},{:0.2f},{:0.2f}])\n'.format(i, *self.dist_restraints[i]['pseudo_atom']['xyz'] * 10)
            for k,serial in enumerate(self.dist_restraints[i]['attach_atoms']['serial']):    
                pymol_vis += 'cmd.distance(\'r{:d}{:d}\', \'p{:d}\', \'id {:d}\')\n'.format(i+1,k+1,i,serial)
        pymol_vis += 'cmd.color(\'gray30\', \'r*\')\n'
        pymol_vis += 'cmd.hide(\'labels\', \'r*\')\n'
        with open(filename, 'w') as f:
            f.write(pymol_vis)


#!/usr/bin/env python3

import pytest
import os
import mdtraj as md
from fretraj import cloud

_TEST_DIR = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))


@pytest.fixture(scope='function')
def create_labels():
    minimal_labels = {"Position":
                      {"Cy3":
                       {"attach_id": 1,
                        "linker_length": 20,
                        "linker_width": 5,
                        "dye_radius1": 8,
                        "dye_radius2": 3,
                        "dye_radius3": 3,
                        "use_LabelLib": False,
                        },
                       "Cy5":
                       {"attach_id": 288,
                        "linker_length": 20,
                        "linker_width": 5,
                        "dye_radius1": 9.5,
                        "dye_radius2": 3,
                        "dye_radius3": 3,
                        "use_LabelLib": False,
                        },
                       },
                      "Distance": {"Cy3-Cy5":
                                   {"R0": 54}
                                   }
                      }
    cloud.check_labels(minimal_labels, verbose=False)
    return minimal_labels


@pytest.fixture(scope='module')
def load_structure():
    return md.load_pdb(os.path.join(_TEST_DIR, 'data/DNA.pdb'))


@pytest.fixture(scope='function')
def create_volumes(create_labels, load_structure):
    vol_Cy3 = cloud.Volume(load_structure, 'Cy3', create_labels)
    vol_Cy5 = cloud.Volume(load_structure, 'Cy5', create_labels)
    return vol_Cy3, vol_Cy5


class TestVolume:

    @pytest.mark.skipif(not cloud._LabelLib_found, reason="requires LabelLib")
    def test_volume_LabelLib(self, create_labels, load_structure):
        create_labels['Position']['Cy3']['use_LabelLib'] = True
        vol = cloud.Volume(load_structure, 'Cy3', create_labels)
        assert hasattr(vol.acv, 'mp')

    def test_volume_pythononly(self, create_labels, load_structure):
        vol = cloud.Volume(load_structure, 'Cy3', create_labels)
        assert hasattr(vol.acv, 'mp')

    def test_stateframe_mismatch(self, create_labels, load_structure):
        create_labels['Position']['Cy3']['state'] = create_labels['Position']['Cy3']['frame_mdtraj']
        vol = cloud.Volume(load_structure, 'Cy3', create_labels)
        assert vol.acv is None

    def test_stateframe_outofrange(self, create_labels, load_structure):
        create_labels['Position']['Cy3']['state'] = 2
        create_labels['Position']['Cy3']['frame_mdtraj'] = 1
        vol = cloud.Volume(load_structure, 'Cy3', create_labels)
        assert vol.acv is None

    def test_attachID_outofrange(self, create_labels, load_structure):
        create_labels['Position']['Cy3']['attach_id'] = 10**3
        vol = cloud.Volume(load_structure, 'Cy3', create_labels)
        assert vol.acv is None

    def test_attachID_buried(self, create_labels, load_structure):
        create_labels['Position']['Cy3']['attach_id'] = 158
        vol = cloud.Volume(load_structure, 'Cy3', create_labels)
        assert vol.acv is None

    def test_fromframes(self, create_labels, load_structure):
        vol = cloud.Volume.from_frames(load_structure, 'Cy3', create_labels, [0, 0])
        assert isinstance(vol[0], cloud.Volume)

    def test_fromattachID(self, create_labels, load_structure):
        vol = cloud.Volume.from_attachID(load_structure, 'Cy3', create_labels, [1, 288])
        assert isinstance(vol[0], cloud.Volume)


class TestFRET:

    def test_FRET_LabelLib(self, create_volumes, create_labels):
        fret = cloud.FRET(create_volumes[0], create_volumes[1], 'Cy3-Cy5', create_labels)
        assert hasattr(fret, 'R_mp')

    def test_FRET_pythononly(self, create_volumes, create_labels):
        create_volumes[0].use_LabelLib = False
        fret = cloud.FRET(create_volumes[0], create_volumes[1], 'Cy3-Cy5', create_labels)
        assert hasattr(fret, 'R_mp')


if __name__ == '__main__':
    pytest.main()



#!/usr/bin/env python3

from pymol import cmd


@cmd.extend
def smooth_map_from_xyz(name, selection, contour_level, grid_spacing, bfactor=100, gaussRes=3, grid_buffer=2):
    """Creates a map object from a selection with xyz coordinates (e.g. a PDB or XYZ object)
    and draws a smooth isosurface at the specified contour level.

    Parameters
    ----------

    name : str
        name of the map
    selection : xyz object
        loaded PDB or XYZ file
    contour_level : float
        contour level (in sigma units)
    grid_spacing : float
        spacing between grid points (in A)
    bfactor : int
        temperature factor; higher numbers generates smoother surfaces
    gaussRes : int
        Gaussian resolution; higher numbers generate smoother surfaces

    Notes
    ----
    If a map for each ACV of a multi-model PDB file should be generated,
    the PDB file must be loaded with the flag discrete=1 (load as discrete objects
    to allow each ACV to have different numbers of grid points.
    (see also https://www.pymolwiki.org/index.php/Discrete_objects)

    """
    name_surf = name + '_isosurf'
    name_map = name + '_map'
    bfactor_str = 'b={:d}'.format(int(bfactor))
    cmd.alter(selection, bfactor_str)
    cmd.alter(selection, bfactor_str)
    gaussRes_default = cmd.get('gaussian_resolution')
    cmd.set('gaussian_resolution', gaussRes)
    # Note: choose state=-3 if all ACVs should be combined into a single map
    cmd.map_new(name_map, 'gaussian', grid_spacing, selection, state=0)
    cmd.isosurface(name_surf, name_map, contour_level, selection, buffer=grid_buffer)
    cmd.set('gaussian_resolution', gaussRes_default)
    cmd.disable(selection)


@cmd.extend
def draw_map(name, isomap, contour_level):
    """
    Draws an isosurface of an open-dx/ccp4 map at the specified contour level.

    Parameters
    ----------
    name : str
        name of the map
    acv_map : map object
        an ccp4 or open-dx map
    contour_level : float
        contour level (in sigma units)
    """
    name_surf = name + '_isosurf'
    cmd.isosurface(name_surf, isomap, contour_level)


@cmd.extend
def set_acv_style(donor_name, acceptor_name, donor_site, acceptor_site, labels, volume_type='AV', transparency=False):
    """
    Set a default style for the ACV clouds

    Parameters
    ----------
    donor_name, acceptor_name : str
        names of the donor and acceptor ACV
    donor_site : str
        reference identifier for the donor labeling position
    acceptor_site : str
        reference identifier for the acceptor labeling position
    labels : dict
        dye, linker and setup parameters for the accessible volume calculation
    volume_type : {'AV', 'CV'}
        entire accessible volume or contact volume
    transparency : bool
        make volume transparent
    """
    cmd.set_color('donor_green', [108, 179, 129])
    cmd.set_color('acceptor_red', [194, 84, 73])
    if volume_type == 'CV':
        contour_level = 'contour_level_CV'
        donor_sele = '{} and resn CV'.format(donor_name)
        acceptor_sele = '{} and resn CV'.format(acceptor_name)
        donor_name = donor_name+'_CV'
        acceptor_name = acceptor_name+'_CV'
    else:
        contour_level = 'contour_level_AV'
        donor_sele = donor_name
        acceptor_sele = acceptor_name
    smooth_map_from_xyz(donor_name, donor_sele, labels['Position'][donor_site][contour_level],
                        labels['Position'][donor_site]['grid_spacing'], labels['Position'][donor_site]['state'])
    smooth_map_from_xyz(acceptor_name, acceptor_sele, labels['Position'][acceptor_site][contour_level],
                        labels['Position'][acceptor_site]['grid_spacing'], labels['Position'][acceptor_site]['state'])
    cmd.color('donor_green', donor_name+'_isosurf')
    cmd.color('acceptor_red', acceptor_name+'_isosurf')
    if transparency:
        cmd.set('transparency', 0.4, donor_name+'_isosurf')
        cmd.set('transparency', 0.4, acceptor_name+'_isosurf')



#!/usr/bin/env python3

import subprocess
import os

package_directory = os.path.dirname(os.path.abspath(__file__))


def pymol_vis():
    subprocess.call(os.path.join(package_directory, 'skripts', 'pymol_vis.sh'))


def vmd_vis():
    subprocess.call(os.path.join(package_directory, 'skripts', 'vmd_vis.sh'))



"""
FRETraj PyMOL Plugin

Calculate accessible contact volumes and predict FRET efficiencies

(c) Fabio Steffen, University of Zurich, 2020-2021
"""

import sys
import os
from PyQt5 import QtWidgets, QtCore, QtGui, uic
import mdtraj as md
import json
import copy
import webbrowser
import re
import functools
import string
import datetime

from fretraj import cloud
from fretraj import metadata

try:
    import LabelLib as ll
except ModuleNotFoundError:
    _LabelLib_found = False
else:
    _LabelLib_found = True

try:
    from pymol import cmd
except ModuleNotFoundError:
    print('Pymol is not installed. Submodule fretraj.isosurf will not be imported.')
else:
    from fretraj import isosurf

package_directory = os.path.dirname(os.path.abspath(cloud.__file__))

dialog = None


def __init_plugin__(app=None):
    """
    Add FRETraj plugin to the Plugins Menu
    """
    from pymol.plugins import addmenuitemqt
    addmenuitemqt('FRETraj', run_plugin_gui)


def run_plugin_gui():
    """
    Create the GUI Window
    """
    global dialog
    if dialog is None:
        dialog = App(_pymol_running=True)
    dialog.show()


class App(QtWidgets.QMainWindow):

    def __init__(self, _pymol_running=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.uiIcon = os.path.join(package_directory, 'UI', 'icon.png')
        self.exampleDataPath = os.path.join(package_directory, 'examples')
        fretrajUI = os.path.join(package_directory, 'UI', 'fretraj.ui')
        self.settingsUI = os.path.join(package_directory, 'UI', 'settings.ui')
        self.textUI = os.path.join(package_directory, 'UI', 'textpad.ui')
        uic.loadUi(fretrajUI, self)
        self.setWindowTitle("FRETraj")
        self.setWindowIcon(QtGui.QIcon(self.uiIcon))
        self._pymol_running = _pymol_running
        self.statusBar().showMessage("Ready", 2000)
        self.readTheDocsURL = None
        self.settingsWindow = QtWidgets.QDialog(self)
        uic.loadUi(self.settingsUI, self.settingsWindow)
        self.settingsWindow.setWindowTitle("FRETraj - Settings")
        self.textWindow = QtWidgets.QDialog(self)
        uic.loadUi(self.textUI, self.textWindow)

        # initialize labels dictionary with defaults from GUI
        self.struct = None
        self.av = {}
        self.traj = {}
        self.labels = {'Position': {}, 'Distance': {}}
        self.labelName = self.comboBox_labelName.currentText()
        self.labelName_default = self.comboBox_labelName.currentText()
        self.distanceName = self.comboBox_distanceName.currentText()
        self.distanceName_default = self.comboBox_distanceName.currentText()
        self.update_labelDict()
        self.labels_default = copy.deepcopy(self.labels)

        # set root path
        root_err = 'No root directory specified. FRETraj is not initialized.'
        self.settings = {'root_path': None, 'browser': None, 'local_docs': None}
        self.settings_file = '{}/.fretraj_settings.json'.format(package_directory)
        if os.path.isfile(self.settings_file):
            with open(self.settings_file, 'r') as f:
                self.settings = json.load(f)
                if os.path.isdir(self.settings['root_path']):
                    self.lineEdit_rootDirectory.setText(self.settings['root_path'])
                    self.settingsWindow.lineEdit_rootDirectory.setText(self.settings['root_path'])
                else:
                    self.settings['root_path'] = None
                    self.setRootDirectory()
                    if not self.settings['root_path']:
                        raise ValueError(root_err)
        else:
            self.setRootDirectory()
            if not self.settings['root_path']:
                raise ValueError(root_err)

        # examples
        n_examples = 0
        fileformat = '.pdb'
        self.exampleQAction = {}
        files = [f for f in os.listdir(self.exampleDataPath) if f.endswith(fileformat)]
        for example in files:
            self.exampleQAction[example] = self.menuLoad_Example.addAction(example.replace(fileformat, ''))
            self.exampleQAction[example].triggered.connect(functools.partial(self.openExample,
                                                           self.exampleQAction[example].text(), fileformat))
            n_examples += 1
        if n_examples > 0:
            self.menuLoad_Example.removeAction(self.example_placeholder)

        # activate / deactivate GUI elements
        self.dyeRadius_spinBox_OnOff()
        self.push_computeACV.setEnabled(False)
        self.doubleSpinBox_contourValue.setEnabled(False)
        self.doubleSpinBox_contourValue_CV.setEnabled(False)
        self.spinBox_bfactor.setEnabled(False)
        self.spinBox_gaussRes.setEnabled(False)
        self.checkBox_transparentAV.setEnabled(False)
        self.doubleSpinBox_gridBuffer.setEnabled(False)
        self.push_calculateFRET.setEnabled(False)
        self.spinBox_statePDB.setEnabled(False)
        self.spinBox_atomID.setEnabled(False)
        self.push_transfer.setEnabled(False)
        self.push_loadParameterFile.setEnabled(False)
        self.push_showText.setEnabled(False)
        self.push_clear.setEnabled(False)
        if not _LabelLib_found:
            self.checkBox_useLabelLib.setEnabled(False)
            self.checkBox_useLabelLib.setChecked(False)

        # signals
        self.push_computeACV.clicked.connect(self.computeACV)
        self.push_loadPDB.clicked.connect(self.loadPDB)
        self.comboBox_simType.activated.connect(self.dyeRadius_spinBox_OnOff)
        self.push_calculateFRET.clicked.connect(self.calculateFRET)
        self.push_deleteLabel.clicked.connect(self.deleteLabel)
        self.push_deleteFRETparam.clicked.connect(self.deleteFRETparam)
        self.push_loadParameterFile.clicked.connect(self.loadParameterFile)
        self.doubleSpinBox_contourValue.valueChanged.connect(self.pymol_update_isosurface)
        self.doubleSpinBox_contourValue_CV.valueChanged.connect(self.pymol_update_isosurface)
        self.spinBox_bfactor.valueChanged.connect(self.pymol_update_isosurface)
        self.spinBox_gaussRes.valueChanged.connect(self.pymol_update_isosurface)
        self.doubleSpinBox_gridBuffer.valueChanged.connect(self.pymol_update_isosurface)
        self.checkBox_transparentAV.toggled.connect(self.pymol_update_isosurface)
        self.lineEdit_labelName.returnPressed.connect(self.makeLabel)
        self.lineEdit_distanceName.returnPressed.connect(self.makeFRETparam)
        self.comboBox_labelName.currentIndexChanged.connect(self.update_comboBox)
        self.actionAbout_FRETraj.triggered.connect(self.openAbout)
        self.comboBox_donorName.currentIndexChanged.connect(self.define_DA)
        self.comboBox_acceptorName.currentIndexChanged.connect(self.define_DA)
        self.comboBox_distanceName.currentIndexChanged.connect(self.update_comboBox)
        self.spinBox_statePDB.valueChanged.connect(self.update_PDBstate)
        self.push_setRootDirectory.clicked.connect(self.setRootDirectory)
        self.push_clear.clicked.connect(self.clear_pymol)
        self.actionDocumentation.triggered.connect(self.openDocumentation)
        self.spinBox_atomID.valueChanged.connect(self.update_atom)
        self.push_transfer.clicked.connect(self.transferToLabel)
        self.push_deleteFRET.clicked.connect(self.deleteFRET)
        self.actionSettings.triggered.connect(self.openSettings)
        self.push_showText.clicked.connect(self.openPDBFile)
        self.settingsWindow.push_root.clicked.connect(self.setRootDirectory)
        self.settingsWindow.push_browser.clicked.connect(self.set_browser)
        self.settingsWindow.push_localdocs.clicked.connect(self.set_localdocsDir)

    def update_labelDict(self, pos=None):
        """
        Update the label dictionary with the values from the GUI fields
        """
        if pos is None:
            pos = self.labelName
        dis = self.distanceName
        self.labels['Position'][pos] = {}
        self.labels['Distance'][dis] = {}
        self.labels['Position'][pos]['attach_id'] = self.spinBox_atomID.value()
        self.labels['Position'][pos]['linker_length'] = self.doubleSpinBox_linkerLength.value()
        self.labels['Position'][pos]['linker_width'] = self.doubleSpinBox_linkerWidth.value()
        self.labels['Position'][pos]['cv_thickness'] = self.doubleSpinBox_CVthickness.value()
        self.labels['Position'][pos]['cv_fraction'] = self.doubleSpinBox_CVfraction.value()
        self.labels['Position'][pos]['simulation_type'] = self.comboBox_simType.currentText()
        self.labels['Position'][pos]['grid_spacing'] = self.doubleSpinBox_gridSpacing.value()
        self.labels['Position'][pos]['dye_radius1'] = self.doubleSpinBox_dyeRadius1.value()
        self.labels['Position'][pos]['dye_radius2'] = self.doubleSpinBox_dyeRadius2.value()
        self.labels['Position'][pos]['dye_radius3'] = self.doubleSpinBox_dyeRadius3.value()
        self.labels['Position'][pos]['mol_selection'] = self.lineEdit_molSelection.text()
        self.labels['Position'][pos]['state'] = self.spinBox_statePDB.value()
        self.labels['Position'][pos]['frame_mdtraj'] = self.spinBox_statePDB.value() - 1
        self.labels['Position'][pos]['use_LabelLib'] = self.checkBox_useLabelLib.isChecked()
        self.labels['Distance'][dis]['R0'] = self.doubleSpinBox_R0.value()
        self.labels['Distance'][dis]['n_dist'] = self.spinBox_nDist.value()
        self.donorName = self.comboBox_donorName.currentText()
        self.acceptorName = self.comboBox_acceptorName.currentText()
        self.labels['Position'][pos]['contour_level_AV'] = self.doubleSpinBox_contourValue.value()
        self.labels['Position'][pos]['contour_level_CV'] = self.doubleSpinBox_contourValue_CV.value()
        self.labels['Position'][pos]['b_factor'] = self.spinBox_bfactor.value()
        self.labels['Position'][pos]['gaussian_resolution'] = self.spinBox_gaussRes.value()
        self.labels['Position'][pos]['grid_buffer'] = self.doubleSpinBox_gridBuffer.value()
        self.labels['Position'][pos]['transparent_AV'] = self.checkBox_transparentAV.isChecked()

    def update_comboBox(self):
        self.update_labelDict()
        self.update_GUIfields()

    def update_GUIfields(self):
        """
        Update the fields of the GUI upon changing the label in the dropdown
        """
        pos = self.comboBox_labelName.currentText()
        dis = self.comboBox_distanceName.currentText()
        self.labelName = pos
        self.distanceName = dis
        self.spinBox_atomID.setValue(self.labels['Position'][pos]['attach_id'])
        self.doubleSpinBox_linkerLength.setValue(self.labels['Position'][pos]['linker_length'])
        self.doubleSpinBox_linkerWidth.setValue(self.labels['Position'][pos]['linker_width'])
        self.doubleSpinBox_CVthickness.setValue(self.labels['Position'][pos]['cv_thickness'])
        self.doubleSpinBox_CVfraction.setValue(self.labels['Position'][pos]['cv_fraction'])
        self.comboBox_simType.setCurrentText(self.labels['Position'][pos]['simulation_type'])
        self.dyeRadius_spinBox_OnOff()
        self.doubleSpinBox_gridSpacing.setValue(self.labels['Position'][pos]['grid_spacing'])
        self.doubleSpinBox_dyeRadius1.setValue(self.labels['Position'][pos]['dye_radius1'])
        self.doubleSpinBox_dyeRadius2.setValue(self.labels['Position'][pos]['dye_radius2'])
        self.doubleSpinBox_dyeRadius3.setValue(self.labels['Position'][pos]['dye_radius3'])
        self.lineEdit_molSelection.setText(self.labels['Position'][pos]['mol_selection'])
        self.spinBox_statePDB.setValue(self.labels['Position'][pos]['state'])
        self.checkBox_useLabelLib.setChecked(self.labels['Position'][pos]['use_LabelLib'])
        self.doubleSpinBox_R0.setValue(self.labels['Distance'][dis]['R0'])
        self.spinBox_nDist.setValue(self.labels['Distance'][dis]['n_dist'])
        # self.update_atom()
        self.doubleSpinBox_contourValue.setValue(self.labels['Position'][pos]['contour_level_AV'])
        self.doubleSpinBox_contourValue_CV.setValue(self.labels['Position'][pos]['contour_level_CV'])
        self.spinBox_bfactor.setValue(self.labels['Position'][pos]['b_factor'])
        self.spinBox_gaussRes.setValue(self.labels['Position'][pos]['gaussian_resolution'])
        self.doubleSpinBox_gridBuffer.setValue(self.labels['Position'][pos]['grid_buffer'])
        self.checkBox_transparentAV.setChecked(self.labels['Position'][pos]['transparent_AV'])

    def loadPDB(self, fileNamePath_pdb=False):
        """
        Load PDB or CIF file. CIF files will be converted to PDB internally
        """
        if not fileNamePath_pdb:
            self.fileNamePath_pdb, _ = QtWidgets.QFileDialog.getOpenFileName(self, 'Load PDB / CIF', '',
                                                                             "PDB / CIF file (*.pdb *cif);;\
                                                                             All Files (*)")
        else:
            self.fileNamePath_pdb = fileNamePath_pdb
        if self.fileNamePath_pdb:
            self.fileName_pdb = self.fileNamePath_pdb.split("/")[-1]
            if self.fileName_pdb[-3:] == 'cif':
                self.fileNamePath_pdb, _ = self.cif2pdb(self.fileNamePath_pdb)
            try:
                self.struct = md.load_pdb(self.fileNamePath_pdb)
            except IndexError:
                self.openErrorWin("File Error", "The specified file \"{}\" cannot be loaded".format(
                    self.fileName_pdb[-3:]))
                return 0
            else:
                self.struct_original = copy.deepcopy(self.struct)
                NA_list = ['A', 'G', 'C', 'U',
                           'RA', 'RG', 'RC', 'RU',
                           'DA', 'DG', 'DC', 'DT',
                           'ATP', 'GTP', 'CTP', 'UTP',
                           'ADP', 'GDP', 'CDP', 'UDP']
                nucleic_str = ' or '.join([f'resn {r}' for r in NA_list])
                idx_protein_nucleic = self.struct.top.select(f'protein or {nucleic_str}')
                self.struct = self.struct.atom_slice(idx_protein_nucleic)

                self.lineEdit_pdbFile.setText(self.fileName_pdb)
                self.spinBox_statePDB.setMaximum(self.struct.n_frames)
                self.spinBox_atomID.setMaximum(self.struct.n_atoms)
                self.atom_chainIDs = [chain.index for chain in self.struct.top.chains for i in range(chain.n_atoms)]
                self.push_computeACV.setEnabled(True)
                self.spinBox_statePDB.setEnabled(True)
                self.spinBox_atomID.setEnabled(True)
                self.push_transfer.setEnabled(True)
                self.push_loadParameterFile.setEnabled(True)
                self.push_clear.setEnabled(True)
                with open(self.fileNamePath_pdb, 'r') as f:
                    self.pdbText = f.read()
                self.push_showText.setEnabled(True)
                if self._pymol_running:
                    cmd.reinitialize()
                    cmd.load(self.fileNamePath_pdb)
                    cmd.remove('solvent or inorganic')
                    chain_startIDs = [1, *[chain.n_atoms+1 for chain in self.struct.top.chains][:-1]]
                    self.chain_names = [cmd.get_chains('index {}'.format(i))[0] for i in chain_startIDs]

                    cmd.set_color('ft_blue', [51, 83, 183])
                    cmd.set_color('ft_gray', [181, 189, 197])
                    cmd.set_color('ft_orange', [227, 128, 82])
                    cmd.hide("nonbonded")
                    cmd.show("cartoon")
                    cmd.cartoon('oval')
                    cmd.set('cartoon_oval_length', 1)
                    cmd.set('cartoon_oval_width', 0.25)
                    cmd.set("cartoon_ring_finder", 2)
                    cmd.set("cartoon_ring_mode", 1)
                    cmd.set("cartoon_ring_transparency", 0.5)
                    cmd.set("cartoon_ring_width", 0.3)
                    cmd.show('sticks', 'name C6+N6+O6+C2+N2+O2+C4+O4+N4 and polymer.nucleic')
                    cmd.set('stick_radius', 0.15, 'polymer.nucleic')
                    cmd.spectrum('count', 'ft_orange ft_gray ft_blue')
                else:
                    self.chain_names = list(string.ascii_uppercase[0:self.struct.top.n_chains])
                self.update_atom()
                return 1

    def cif2pdb(self, pathtoPDBfile):
        """
        Convert .cif to .pdb with PyMOL

        Returns
        -------
        fileNamePath_pdb : string
        fileName_pdb : string
        """
        name = pathtoPDBfile.split("/")[-1][:-4]
        try:
            cmd.load(pathtoPDBfile)
        except:
            self.openErrorWin("PDB File Error", "The specified file \"{}.pdb\" can not be loaded".format(name))
        else:
            fileNamePath_pdb = '{}/{}.pdb'.format(self.settings['root_path'], name)
            fileName_pdb = '{}.pdb'.format(fileNamePath_pdb.split("/")[-1])
            cmd.save(fileNamePath_pdb, name)
        return fileNamePath_pdb, fileName_pdb

    def update_PDBstate(self):
        if self._pymol_running:
            self.update_labelDict()
            cmd.set('state', self.labels['Position'][self.labelName]['state'])

    def update_atom(self):
        atom_id = self.spinBox_atomID.value() - 1
        chain_id = self.atom_chainIDs[atom_id]
        self.lineEdit_pdbAtom.setText('{}-{}'.format(self.chain_names[chain_id], str(self.struct.top.atom(atom_id))))

    def loadParameterFile(self, fileNamePath_param=False):
        """
        Load dye, linker and simulation settings from parameter file
        """
        if not fileNamePath_param:
            self.fileNamePath_param, _ = QtWidgets.QFileDialog.getOpenFileName(self, 'Load label parameter file', '',
                                                                               "JSON Files (*.json);;All Files (*)")
        else:
            self.fileNamePath_param = fileNamePath_param
        if self.fileNamePath_param:
            self.fileName_param = self.fileNamePath_param.split("/")[-1]
            try:
                with open(self.fileNamePath_param) as f:
                    try:
                        labels_json = json.load(f)
                    except json.decoder.JSONDecodeError:
                        self.openErrorWin('Parameter File Error',
                                          'The specified file \"{}\" is not of type JSON'.format(self.fileName_param))
                        return 0
                    else:
                        error_msg = "The specified file \"{}\" has a wrong format".format(self.fileName_param)
                        for field in labels_json.keys():
                            if field == 'Distance':
                                name_default = self.distanceName_default
                            else:
                                name_default = self.labelName_default

                            if not isinstance(labels_json[field], dict):
                                self.openErrorWin('Parameter File Error', error_msg)
                                return 0

                            for pos_dis in labels_json[field].keys():

                                self.labels[field][pos_dis] = {}
                                for key in self.labels_default[field][name_default].keys():
                                    if isinstance(labels_json[field][pos_dis], dict):
                                        if key in labels_json[field][pos_dis].keys():
                                            self.labels[field][pos_dis][key] = labels_json[field][pos_dis][key]
                                        else:
                                            self.labels[field][pos_dis][key] = copy.deepcopy(
                                                self.labels_default[field][name_default][key])
                                    else:
                                        self.openErrorWin('Parameter File Error', error_msg)
                                        return 0
                        self.lineEdit_paramFile.setText(self.fileName_param)
                    for newlabel in self.labels['Position'].keys():
                        self.addLabel(newlabel)
                    for newdistance in self.labels['Distance'].keys():
                        self.addFRETparam(newdistance)
                    return 1
            except FileNotFoundError:
                self.openErrorWin('Parameter File Error',
                                  'The specified file \"{}\" could not be found'.format(self.fileName_param))
                return 0

    def transferToLabel(self):
        newlabel = '{:d}-{}'.format(self.spinBox_statePDB.value(), self.lineEdit_pdbAtom.text())
        self.labels['Position'][newlabel] = copy.deepcopy(self.labels['Position'][self.labelName])
        self.update_labelDict(newlabel)
        self.update_GUIfields()
        self.addLabel(newlabel)

    def makeLabel(self):
        """
        Create new label from edit box string
        """
        newlabel = self.lineEdit_labelName.text()
        self.labels['Position'][newlabel] = copy.deepcopy(self.labels['Position'][self.labelName])
        self.update_labelDict(newlabel)
        self.update_GUIfields()
        self.addLabel(newlabel)

    def addLabel(self, newlabel):
        """
        Add a new label to Combobox
        """
        if self.comboBox_labelName.findText(newlabel) == -1:
            self.comboBox_labelName.addItem(newlabel)
            self.comboBox_labelName.setCurrentText(newlabel)

        self.lineEdit_labelName.clear()
        if self.struct:
            self.push_computeACV.setEnabled(True)
        if self.tableWidget_MeanPos.rowCount() > 1 and self.struct:
            self.push_calculateFRET.setEnabled(True)
        self.update_GUIfields()

    def deleteLabel(self):
        """
        Remove the selected label from Combobox
        """
        self.deleteLabelFromList()
        self.deleteDistanceFromList()
        self.deleteIsosurface()
        if self.labelName != self.labelName_default:
            todelete = self.labelName
            self.comboBox_labelName.removeItem(self.comboBox_labelName.currentIndex())
            self.lineEdit_labelName.clear()
            self.labels['Position'].pop(todelete)

    def addLabelToList(self, av):
        """
        Add a label and the computed mean dye position (MP) to the list
        """
        rowCount = self.tableWidget_MeanPos.rowCount()
        for r in range(rowCount):
            if self.labelName == self.tableWidget_MeanPos.item(r, 0).text():
                break
        else:
            r = 0
            self.tableWidget_MeanPos.insertRow(r)
        self.tableWidget_MeanPos.setItem(r, 0, QtWidgets.QTableWidgetItem(self.labelName))
        self.tableWidget_MeanPos.setItem(r, 1, QtWidgets.QTableWidgetItem(', '.join(
            '{:.2f}'.format(p) for p in av.acv.mp)))
        if self.comboBox_donorName.findText(self.labelName) == -1:
            self.comboBox_donorName.addItem(self.labelName)
        if self.comboBox_acceptorName.findText(self.labelName) == -1:
            self.comboBox_acceptorName.addItem(self.labelName)

    def deleteLabelFromList(self):
        """
        Remove a label and the computed mean dye position (MP) from the list
        """
        r = 0
        rowCount = self.tableWidget_MeanPos.rowCount()
        while r < rowCount:
            if self.labelName in self.tableWidget_MeanPos.item(r, 0).text():
                self.tableWidget_MeanPos.removeRow(r)
                rowCount -= 1
            else:
                r += 1

        i = 0
        itemCount = self.comboBox_donorName.count()
        while i < itemCount:
            if self.labelName in self.comboBox_donorName.itemText(i):
                self.comboBox_donorName.removeItem(i)
                self.comboBox_acceptorName.removeItem(i)
                itemCount -= 1
            else:
                i += 1

    def deleteIsosurface(self):
        av_name = self.fileName_pdb[:-4]+'-'+self.labelName.replace('\'', 'p')
        if av_name in cmd.get_names('objects'):
            cmd.delete(av_name)
            cmd.delete(av_name + '_map')
            cmd.delete(av_name + '_isosurf')
            cmd.delete(av_name + '_CV_map')
            cmd.delete(av_name + '_CV_isosurf')

    def makeFRETparam(self):
        """
        Create new FRET parameters from edit box string
        """
        newdistance = self.lineEdit_distanceName.text()
        self.labels['Distance'][newdistance] = copy.deepcopy(self.labels_default['Distance'][self.distanceName_default])
        self.addFRETparam(newdistance)

    def addFRETparam(self, newdistance):
        """
        Add FRET parameters to the Combobox
        """
        if self.comboBox_distanceName.findText(newdistance) == -1:
            self.comboBox_distanceName.addItem(newdistance)
            self.comboBox_distanceName.setCurrentText(newdistance)
        self.lineEdit_distanceName.clear()

    def deleteFRETparam(self):
        """
        Remove FRET parameters from the Combobox
        """
        dis = self.comboBox_distanceName.currentText()
        if dis != self.distanceName_default:
            self.comboBox_distanceName.removeItem(self.comboBox_distanceName.currentIndex())

    def addDistanceToList(self, traj):
        """
        Add a FRET set to the list
        """
        DA = '{} -> {}'.format(self.donorName, self.acceptorName)
        rowCount = self.tableWidget_FRET.rowCount()
        for r in range(rowCount):
            if DA == self.tableWidget_FRET.item(r, 0).text():
                break
        else:
            r = 0
            self.tableWidget_FRET.insertRow(r)
        self.tableWidget_FRET.setItem(r, 0, QtWidgets.QTableWidgetItem('{} -> {}'.format(self.donorName,
                                                                                         self.acceptorName)))
        self.tableWidget_FRET.setItem(r, 1, QtWidgets.QTableWidgetItem('{:.1f}'.format(traj.mean_R_DA)))
        self.tableWidget_FRET.setItem(r, 2, QtWidgets.QTableWidgetItem('{:.1f}'.format(traj.sigma_R_DA)))
        self.tableWidget_FRET.setItem(r, 3, QtWidgets.QTableWidgetItem('{:.2f}'.format(traj.mean_E_DA)))
        self.tableWidget_FRET.setItem(r, 4, QtWidgets.QTableWidgetItem('{:.1f}'.format(traj.mean_R_DA_E)))
        self.tableWidget_FRET.setItem(r, 5, QtWidgets.QTableWidgetItem('{:.1f}'.format(traj.R_mp)))
        self.tableWidget_FRET.setItem(r, 6, QtWidgets.QTableWidgetItem('{:.1f}'.format(traj.R_attach)))

    def deleteDistanceFromList(self):
        """
        Remove a FRET set from the list
        """
        r = 0
        rowCount = self.tableWidget_FRET.rowCount()
        while r < rowCount:
            if self.labelName in self.tableWidget_FRET.item(r, 0).text():
                self.tableWidget_FRET.removeRow(r)
                rowCount -= 1
            else:
                r += 1

    def define_DA(self):
        """
        Define the donor and acceptor labels and color them in the mean position and in the FRET table
        """
        self.update_labelDict()
        for i in range(self.tableWidget_MeanPos.rowCount()):
            for j in range(self.tableWidget_MeanPos.columnCount()):
                self.tableWidget_MeanPos.item(i, j).setBackground(QtGui.QColor(255, 255, 255))

        for i in range(self.tableWidget_FRET.rowCount()):
            for j in range(self.tableWidget_FRET.columnCount()):
                self.tableWidget_FRET.item(i, j).setBackground(QtGui.QColor(255, 255, 255))

        if self.tableWidget_MeanPos.rowCount() > 1:
            if self.donorName == self.acceptorName:
                self.push_calculateFRET.setEnabled(False)
                if self._pymol_running:
                    cmd.color('white', '*_isosurf')
            else:
                self.push_calculateFRET.setEnabled(True)
                row_don = self.tableWidget_MeanPos.findItems(self.donorName, QtCore.Qt.MatchExactly)[0].row()
                row_acc = self.tableWidget_MeanPos.findItems(self.acceptorName, QtCore.Qt.MatchExactly)[0].row()
                for j in range(self.tableWidget_MeanPos.columnCount()):
                    self.tableWidget_MeanPos.item(row_don, j).setBackground(QtGui.QColor(108, 179, 129))
                    self.tableWidget_MeanPos.item(row_acc, j).setBackground(QtGui.QColor(194, 84, 73))

                DA = '{} -> {}'.format(self.donorName, self.acceptorName)
                for r in range(self.tableWidget_FRET.rowCount()):
                    if DA in self.tableWidget_FRET.item(r, 0).text():
                        for j in range(self.tableWidget_FRET.columnCount()):
                            self.tableWidget_FRET.item(r, j).setBackground(QtGui.QColor(200, 200, 200))
                        break

                if self._pymol_running:
                    cmd.color('white', '*_isosurf')
                    don_name = self.fileName_pdb[:-4]+'-'+self.donorName.replace('\'', 'p')
                    acc_name = self.fileName_pdb[:-4]+'-'+self.acceptorName.replace('\'', 'p')
                    cmd.set_color('don_green', [108, 179, 129])
                    cmd.set_color('acc_red', [194, 84, 73])
                    cmd.color('don_green', don_name + '_isosurf')
                    cmd.color('acc_red', acc_name + '_isosurf')
                    if any(self.av[self.donorName].acv.tag_1d > 1):
                        cmd.color('don_green', don_name + '_CV_isosurf')
                    if any(self.av[self.acceptorName].acv.tag_1d > 1):
                        cmd.color('acc_red', acc_name + '_CV_isosurf')

        else:
            self.push_calculateFRET.setEnabled(False)

    def dyeRadius_spinBox_OnOff(self):
        """
        Activate or deactivate dyeRadius spinboxes depending on the type of simulation
        """
        if self.comboBox_simType.currentText() == 'AV1':
            self.doubleSpinBox_dyeRadius2.setEnabled(False)
            self.doubleSpinBox_dyeRadius3.setEnabled(False)
        else:
            self.doubleSpinBox_dyeRadius2.setEnabled(True)
            self.doubleSpinBox_dyeRadius3.setEnabled(True)

    def computeACV(self):
        """
        Create a cloud.Volume object at the specified label position.
        If the GUI is run as a PyMOL plugin, the ACV will be rendered in the current window
        """
        self.update_labelDict()
        msg = 'Busy...'
        self.statusBar().showMessage(msg, 3000)
        param_filename = '{}/{}_parameters.json'.format(self.settings['root_path'], self.fileName_pdb[:-4])
        cloud.save_labels(param_filename, self.labels)
        self.av[self.labelName] = cloud.Volume(self.struct, self.labelName, self.labels)
        if self.av[self.labelName].acv is None:
            msg = 'ACV could not be calculated!'
            self.statusBar().showMessage(msg, 3000)
            print(msg)
        else:
            av_name = self.fileName_pdb[:-4]+'-'+self.labelName.replace('\'', 'p')
            av_filename = '{}/{}.pdb'.format(self.settings['root_path'], av_name)
            self.av[self.labelName].save_acv(av_filename, format='pdb')

            self.addLabelToList(self.av[self.labelName])
            msg = 'ACV successfully calculated!'
            self.statusBar().showMessage(msg, 3000)
            if self._pymol_running:
                cmd.delete('{}*'.format(av_name))
                cmd.load(av_filename)
                self.doubleSpinBox_contourValue.setEnabled(True)
                self.spinBox_bfactor.setEnabled(True)
                self.spinBox_gaussRes.setEnabled(True)
                self.doubleSpinBox_gridBuffer.setEnabled(True)
                self.checkBox_transparentAV.setEnabled(True)
                self.pymol_update_isosurface()
                cmd.zoom(self.fileName_pdb[:-4])
            self.define_DA()

    def calculateFRET(self):
        self.update_labelDict()
        self.traj[(self.donorName, self.acceptorName)] = cloud.FRET(self.av[self.donorName], self.av[self.acceptorName],
                                                                    self.distanceName, self.labels)
        self.addDistanceToList(self.traj[(self.donorName, self.acceptorName)])
        self.define_DA()
        self.traj[(self.donorName, self.acceptorName)].save_fret('{}/{}_{}_{}_fret.json'.format(
            self.settings['root_path'], self.fileName_pdb[:-4], self.donorName, self.acceptorName))

    def deleteFRET(self):
        DA = '{} -> {}'.format(self.donorName, self.acceptorName)
        r = 0
        rowCount = self.tableWidget_FRET.rowCount()
        while r < rowCount:
            if DA in self.tableWidget_FRET.item(r, 0).text():
                self.tableWidget_FRET.removeRow(r)
                rowCount -= 1
            else:
                r += 1

    def pymol_update_isosurface(self):
        """
        Update the isosurface when spin box value is changed.
        The spin box is only active when PYMOL is running.
        """
        if self.av:
            av_name = self.fileName_pdb[:-4]+'-'+self.labelName.replace('\'', 'p')
            contour_level = self.doubleSpinBox_contourValue.value()
            bfactor = self.spinBox_bfactor.value()
            gaussRes = self.spinBox_gaussRes.value()
            gridBuffer = self.doubleSpinBox_gridBuffer.value()
            grid_spacing = self.labels['Position'][self.labelName]['grid_spacing']
            if av_name in cmd.get_names('objects'):
                isosurf.smooth_map_from_xyz(av_name, av_name, contour_level, grid_spacing, bfactor,
                                            gaussRes, gridBuffer)
                if any(self.av[self.labelName].acv.tag_1d > 1):
                    self.doubleSpinBox_contourValue_CV.setEnabled(True)
                    contour_level_CV = self.doubleSpinBox_contourValue_CV.value()
                    sele_CV = '{} and resn CV'.format(av_name)
                    isosurf.smooth_map_from_xyz(av_name+'_CV', sele_CV, contour_level_CV, grid_spacing, bfactor,
                                                gaussRes, gridBuffer)
                    if self.checkBox_transparentAV.isChecked():
                        cmd.set('transparency', 0.4, av_name + '_isosurf')
                    else:
                        cmd.set('transparency', 0, av_name + '_isosurf')
                else:
                    self.doubleSpinBox_contourValue_CV.setEnabled(False)
                    if self.checkBox_transparentAV.isChecked():
                        cmd.set('transparency', 0.4, av_name + '_isosurf')
                    else:
                        cmd.set('transparency', 0, av_name + '_isosurf')

    def clear_pymol(self):
        self.lineEdit_pdbFile.clear()
        self.lineEdit_paramFile.clear()
        self.push_computeACV.setEnabled(False)
        self.spinBox_statePDB.setValue(1)
        self.spinBox_atomID.setValue(1)
        self.spinBox_statePDB.setEnabled(False)
        self.spinBox_atomID.setEnabled(False)
        self.push_transfer.setEnabled(False)
        self.push_loadParameterFile.setEnabled(False)
        self.push_showText.setEnabled(False)
        self.lineEdit_pdbAtom.clear()
        for _ in range(self.comboBox_labelName.count()):
            self.deleteLabel()
        cmd.reinitialize()

    def setRootDirectory(self):
        rootDir = QtWidgets.QFileDialog.getExistingDirectory(self, 'Set root directory')
        if rootDir:
            self.lineEdit_rootDirectory.setText(rootDir)
            self.settingsWindow.lineEdit_rootDirectory.setText(rootDir)
            self.settings['root_path'] = re.escape(rootDir)
            with open(self.settings_file, 'w') as f:
                json.dump(self.settings, f, indent=2)

    def set_browser(self):
        browser_path, _ = QtWidgets.QFileDialog.getOpenFileName(self, 'Search for default browser')
        if browser_path:
            self.settingsWindow.lineEdit_browser.setText(browser_path)
            self.settings['browser'] = re.escape(browser_path)
            with open(self.settings_file, 'w') as f:
                json.dump(self.settings, f, indent=2)

    def set_localdocsDir(self):
        docs_path = QtWidgets.QFileDialog.getExistingDirectory(self, 'Select docs directory')
        if docs_path:
            self.settingsWindow.lineEdit_localdocs.setText(docs_path)
            self.settings['local_docs'] = re.escape(docs_path)
            with open(self.settings_file, 'w') as f:
                json.dump(self.settings, f, indent=2)

    def openDocumentation(self):
        if not self.readTheDocsURL:
            if not self.settings['local_docs']:
                msg = QtWidgets.QMessageBox()
                msg.setIcon(QtWidgets.QMessageBox.Information)
                msg.setWindowTitle("Location of docs not configured")
                msg.setStandardButtons(QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel)
                msg.setText('Press <OK> and specify the path of the local docs.')
                returnValue = msg.exec_()
                if returnValue == QtWidgets.QMessageBox.Ok:
                    self.set_localdocsDir()

        if not self.settings['browser']:
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Information)
            msg.setWindowTitle("Web browser not configured")
            msg.setStandardButtons(QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel)
            msg.setText('For the documentation to be displayed, the path to a web browser needs to be configured. \
                         Press <OK> and search for the browser executable (Firefox, Chrome or Edge).')
            returnValue = msg.exec_()
            if returnValue == QtWidgets.QMessageBox.Ok:
                self.set_browser()

        if self.settings['browser']:
            try:
                browser = webbrowser.get('{} %s'.format(self.settings['browser']))
                browser.open('file://{}/index.html'.format(self.settings['local_docs']))
            except webbrowser.Error:
                self.settings['browser'] = None
                print('Browser not found!')
                with open(self.settings_file, 'w') as f:
                    json.dump(self.settings, f, indent=2)
            except FileNotFoundError:
                self.settings['local_docs'] = None
                print('Local docs not found!')
                with open(self.settings_file, 'w') as f:
                    json.dump(self.settings, f, indent=2)

    def openAbout(self):
        """
        Open About Window
        """
        msg = QtWidgets.QMessageBox()
        pixmap = QtGui.QPixmap(self.uiIcon)
        msg.setIconPixmap(pixmap.scaledToWidth(64))
        msg.setWindowTitle("About FRETraj")
        current_year = datetime.datetime.now().year
        msg.setText(f"{metadata['Name']} {metadata['Version']}\n{metadata['Summary']}\n\n\
                     (C) {metadata['Author']}\nUniversity of Zurich, 2020-{current_year}")
        msg.exec_()

    def openExample(self, name, fileformat):
        """
        Load an example file and calculate an ACV
        """
        fileNamePath_pdb = '{}/{}{}'.format(self.exampleDataPath, name, fileformat)
        fileNamePath_param = '{}/{}_labels{}'.format(self.exampleDataPath, name, '.json')
        pdb_load = self.loadPDB(fileNamePath_pdb)
        if pdb_load:
            param_load = self.loadParameterFile(fileNamePath_param)
            if param_load:
                msg = QtWidgets.QMessageBox()
                msg.setIcon(QtWidgets.QMessageBox.Information)
                msg.setWindowTitle("Calculate ACV?")
                msg.setStandardButtons(QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
                msg.setText('Would you like to calculate an ACV?')
                returnValue = msg.exec_()
                if returnValue == QtWidgets.QMessageBox.Yes:
                    n_pos = 0
                    for pos in self.labels['Position'].keys():
                        if pos != self.labelName_default:
                            index = self.comboBox_labelName.findText(pos, QtCore.Qt.MatchExactly)
                            self.comboBox_labelName.setCurrentIndex(index)
                            self.computeACV()
                            n_pos += 1
                    if n_pos > 1:
                        self.comboBox_donorName.setStyleSheet('QComboBox {background-color: rgb(229,134,145);}')
                        self.comboBox_acceptorName.setStyleSheet('QComboBox {background-color: rgb(229,134,145);}')
                        msg = QtWidgets.QMessageBox()
                        msg.setIcon(QtWidgets.QMessageBox.Information)
                        msg.setWindowTitle("Calculate FRET?")
                        msg.setText('To calculate FRET select the donor and acceptor position \
                                     from the highlighted dropdown menu and press on <Calculate FRET>')
                        returnValue = msg.exec_()
                        self.comboBox_donorName.setStyleSheet('')
                        self.comboBox_acceptorName.setStyleSheet('')

    def openSettings(self):
        self.settingsWindow.lineEdit_rootDirectory.setText(self.settings['root_path'])
        self.settingsWindow.lineEdit_browser.setText(self.settings['browser'])
        self.settingsWindow.lineEdit_localdocs.setText(self.settings['local_docs'])
        isOK = self.settingsWindow.exec_()
        if isOK:
            self.settings['root_path'] = self.settingsWindow.lineEdit_rootDirectory.text()
            self.settings['browser'] = self.settingsWindow.lineEdit_browser.text()
            self.settings['local_docs'] = self.settingsWindow.lineEdit_localdocs.text()
            self.lineEdit_rootDirectory.setText(self.settings['root_path'])
            with open(self.settings_file, 'w') as f:
                json.dump(self.settings, f, indent=2)

    def openPDBFile(self):
        self.textWindow.textBrowser_pdbFile.setText(self.pdbText)
        self.textWindow.setWindowTitle("FRETraj - {}".format(self.fileName_pdb))
        self.textWindow.exec_()

    def openErrorWin(self, title, message):
        """
        Open Error Window
        """
        msg = QtWidgets.QMessageBox()
        msg.setIcon(QtWidgets.QMessageBox.Critical)
        msg.setWindowTitle(title)
        msg.setText(message)
        msg.exec_()


def main():
    app = QtWidgets.QApplication(sys.argv)
    window = App()
    window.show()
    app.exec_()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# start a PyMOL server session from a terminal:
#   pymol -R

# On Windows you may create a shortcut that executes the following command:
#   C:\path\to\PyMOLWin.exe -R

import os
import re
import nglview
import ipywidgets


def connect2pymol():
    """
    Establish an RPC connection to run PyMOL commands from the command line or from a Jupyter notebook
    """
    import xmlrpc.client as xmlrpclib
    cmd = xmlrpclib.ServerProxy('http://localhost:9123')
    curr_wd = os.getcwd()
    try:
        cmd.cd(curr_wd)
    except:
        cmd.cd(re.sub(r'/mnt/([a-z])', r'\1:', curr_wd))
    return cmd


def nglview_trajectory(traj_biomol):
    view = nglview.NGLWidget()
    view.add_trajectory(traj_biomol)
    view.clear_representations(component=0)
    view.add_simplified_base(component=0, selection='/0', disablePicking=True, colorScheme='atomindex')
    view.add_cartoon(component=0, selection='/0', aspectRatio=4, disablePicking=True, colorScheme='atomindex')
    view.stage.set_parameters(mouse_preset='pymol')
    return view


def nglview_trajectory_AV(traj_biomol, traj_volume1, traj_volume2, surface_representation=False):
    """Create a nglview trajectory scene with donor and acceptor accessible volumes (AV)

    Parameters
    ----------
    traj_biomol : mdtraj.Trajectory
        trajectory of the biomolecule
    traj_volume1, traj_volume2 : mdtraj.Trajectory
        trajectory of the donor and acceptor accessible volume
    surface_representation : bool (default: False)
        show a surface representation (instead of spacefill, which is faster)

    Returns
    -------
    view : nglview.NGLWidget
    """
    view = nglview.NGLWidget()
    view.add_trajectory(traj_biomol)
    view.add_trajectory(traj_volume1)
    view.add_trajectory(traj_volume2)

    view.clear_representations(component=0)
    view.clear_representations(component=1)
    view.clear_representations(component=2)

    view.add_simplified_base(component=0, selection='/0', disablePicking=True, colorScheme='atomindex')
    view.add_cartoon(component=0, selection='/0', aspectRatio=4, disablePicking=True, colorScheme='atomindex')

    if surface_representation:
        view.add_surface(color='#6cb381', wireframe=True, opacity=0.4, isolevel=0, component=1,
                         disablePicking=True, selection='all')
        view.add_surface(color='#c25449', wireframe=True, opacity=0.4, isolevel=0, component=2,
                         disablePicking=True, selection='all')
    else:
        view.add_spacefill(color='#6cb381', component=1, disablePicking=True, selection='all')
        view.add_spacefill(color='#c25449', component=2, disablePicking=True, selection='all')
    view.stage.set_parameters(mouse_preset='pymol')
    return view


def nglview_trajectory_ACV(traj_biomol, traj_volume1_FV, traj_volume2_FV, traj_volume1_CV, traj_volume2_CV):
    """Create a nglview trajectory scene with donor and acceptor accessible-contact volumes (ACV)

    Parameters
    ----------
    traj_biomol : mdtraj.Trajectory
        trajectory of the biomolecule
    traj_volume1_FV, traj_volume2_FV : mdtraj.Trajectory
        trajectory of the donor and acceptor free volume
    traj_volume1_CV, traj_volume2_CV : mdtraj.Trajectory
        trajectory of the donor and acceptor contact volume

    Returns
    -------
    view : nglview.NGLWidget
    """
    view = nglview.NGLWidget()
    view.add_trajectory(traj_biomol)
    view.add_trajectory(traj_volume1_CV)
    view.add_trajectory(traj_volume2_CV)
    view.add_trajectory(traj_volume1_FV)
    view.add_trajectory(traj_volume2_FV)

    view.clear_representations(component=0)
    view.clear_representations(component=1)
    view.clear_representations(component=2)
    view.clear_representations(component=3)
    view.clear_representations(component=4)

    view.add_simplified_base(component=0, selection='/0', disablePicking=True, colorScheme='atomindex')
    view.add_cartoon(component=0, selection='/0', aspectRatio=4, disablePicking=True, colorScheme='atomindex')

    view.add_surface(color='#6cb381', wireframe=False, opacity=0.4, isolevel=0, component=1,
                     disablePicking=True, selection='all')
    view.add_surface(color='#c25449', wireframe=False, opacity=0.4, isolevel=0, component=2,
                     disablePicking=True, selection='all')
    view.add_surface(color='#6cb381', wireframe=True, opacity=0.4, isolevel=0, component=3,
                     disablePicking=True, selection='all')
    view.add_surface(color='#c25449', wireframe=True, opacity=0.4, isolevel=0, component=4,
                     disablePicking=True, selection='all')
    view.stage.set_parameters(mouse_preset='pymol')
    return view


def nglview_multimodel_ACV(biomol_filename, volume1_filename, volume2_filename):
    """Create a nglview multi-model structure scene with donor and acceptor accessible-contact volumes (ACV)

    Parameters
    ----------
    structure_filename : str
                         multi-model PDB of the biomolecule
    volume1_filename, volume2_filename : str
        multi-model PDB of the donor and acceptor accessible-contact volume

    Returns
    -------
    view : nglview.NGLWidget
    n_model : int
    """
    global view
    view = nglview.NGLWidget()
    struct = nglview.FileStructure(biomol_filename)
    acv_D = nglview.FileStructure(volume1_filename)
    acv_A = nglview.FileStructure(volume2_filename)

    struct_str = struct.get_structure_string()
    n_models = struct_str.count('MODEL')

    view.add_component(struct)
    view.add_component(acv_D)
    view.add_component(acv_A)

    view.clear_representations(component=0)
    view.clear_representations(component=1)
    view.clear_representations(component=2)

    view.add_simplified_base(component=0, selection='/0', disablePicking=True, colorScheme='atomindex')
    view.add_cartoon(component=0, selection='/0', aspectRatio=4, disablePicking=True, colorScheme='atomindex')

    view.add_surface(color='#6cb381', wireframe=False, opacity=0.4, isolevel=0, component=1,
                     disablePicking=True, selection='CV')
    view.add_surface(color='#c25449', wireframe=False, opacity=0.4, isolevel=0, component=2,
                     disablePicking=True, selection='CV')
    view.add_surface(color='#6cb381', wireframe=True, opacity=0.4, isolevel=0, component=1,
                     disablePicking=True, selection='all')
    view.add_surface(color='#c25449', wireframe=True, opacity=0.4, isolevel=0, component=2,
                     disablePicking=True, selection='all')
    view.stage.set_parameters(mouse_preset='pymol')
    return view, n_models


def _change_model(model):
    """Change the current model

    Parameters
    ----------
    model : int
    """
    for i in range(view.n_components):
        view._remote_call('setSelection', target='compList', args=[f'/{model}'], kwargs=dict(component_index=i))
        n_representations = 5  # more than 5 representations are unlikely
        for i in range(n_representations):
            view._remote_call('setSelection', target='Representation', args=[f'/{model}'],
                              kwargs=dict(component_index=0, repr_index=i))


def model_slider(n_models):
    """A slider widget for multi-model structures

    Parameters
    ----------
    n_models : int
               number of models in the structure

    Returns
    -------
    int_slider: ipywidgets.IntSlider

    Notes
    -----
    See also: https://ipywidgets.readthedocs.io/en/stable/examples/Widget%20List.html
    """
    int_slider = ipywidgets.IntSlider(
        value=0,
        min=0,
        max=n_models-1,
        step=1,
        description='Frame:',
        disabled=False,
        continuous_update=False,
        orientation='horizontal',
        readout=True,
        readout_format='d'
        )
    int_slider.observe(_on_slider_value_change, names='value')
    return int_slider


def _on_slider_value_change(change):
    """Updates the view upon changing the slider value

    Parameters
    ----------
    change : dict

    Notes
    -----
    See also: https://ipywidgets.readthedocs.io/en/latest/examples/Widget%20Events.html
    """
    output = ipywidgets.Output()
    with output:
        _change_model(change['new'])


def render_view(view, gui=False):
    """Display an nglview scene with or without a GUI

    Parameters
    ----------
    view : nglview.NGLWidget
    gui : bool
          display the scene within a GUI window
          (useful for interacting with components and representations)
    """
    print('Building scene, please wait...')
    return view.display(gui)



#!/usr/bin/env python3

import numpy as np
try:
    import LabelLib as ll
except ModuleNotFoundError:
    _LabelLib_found = False
else:
    _LabelLib_found = True


def dist_mp(acv1, acv2):
    """Compute the distance between mean dye positions R_MP of two accessible 
    contact volumes

    Parameters
    ----------
    acv1, acv2 : fretraj.cloud.ACV
        donor and acceptor ACVs

    Returns
    -------
    float
        distance between the mean dye positions of acv1 and acv2
    """
    R_mp = np.sqrt(sum((acv1.mp - acv2.mp)**2))
    return R_mp


def dists_DA(acv1, acv2, n_dist=10**6, return_weights=True):
    """Compute a randomly subsampled donor-acceptor distance distribution
    for the two accessible volume clouds

    Parameters
    ----------
    acv1, acv2 : fretraj.cloud.ACV
        donor and acceptor ACVs
    n_dist : int, default=10E6
        number of distances to calculate
    return_weights : bool, default=True
        additional vector with weights

    Returns
    -------
    ndarray
        ndarray of shape 1 x n_dist or 2 x n_dist (if weights are included)
        containing distances between acv1 and acv2
    """
    if return_weights:
        idx1 = np.random.choice(len(acv1.cloud_xyzqt), n_dist)
        idx2 = np.random.choice(len(acv2.cloud_xyzqt), n_dist)
        R_DA = np.sqrt(np.sum((acv1.cloud_xyzqt[idx1, 0:3]
                               - acv2.cloud_xyzqt[idx2, 0:3])**2, axis=1))
        w = acv1.cloud_xyzqt[idx1, 3] * acv2.cloud_xyzqt[idx2, 3]
        w = w / w.sum()
        R_DA = np.vstack((R_DA, w)).T
    else:
        w1 = acv1.cloud_xyzqt[:, 3] / sum(acv1.cloud_xyzqt[:, 3])
        w2 = acv2.cloud_xyzqt[:, 3] / sum(acv2.cloud_xyzqt[:, 3])
        idx1 = np.random.choice(len(acv1.cloud_xyzqt), n_dist, p=w1)
        idx2 = np.random.choice(len(acv2.cloud_xyzqt), n_dist, p=w2)
        R_DA = np.sqrt(np.sum((acv1.cloud_xyzqt[idx1, 0:3]
                               - acv2.cloud_xyzqt[idx2, 0:3])**2, axis=1))
    return R_DA


def dists_DA_ll(acv1, acv2, n_dist=10**6, return_weights=True):
    """Compute a randomly subsampled donor-acceptor distance distribution
    for the two accessible volume clouds

    Parameters
    ----------
    acv1, acv2 : fretraj.cloud.ACV
        donor and acceptor ACVs
    n_dist : int, default=10E6
        number of distances to calculate
    include_weights : bool, default=True
        additional vector with weights

    Returns
    -------
    ndarray
        ndarray of shape 1 x n_dist or 2 x n_dist (if weights are included)
        containing distances between acv1 and acv2
    """
    R_DA = np.array(ll.sampleDistanceDistInv(acv1.ll_Grid3D, acv2.ll_Grid3D,
                    n_dist))
    if return_weights:
        w = np.full((1, n_dist), 1 / n_dist)
        R_DA = np.vstack((R_DA, w)).T
    return R_DA


def mean_dist_DA(acv1, acv2, R_DA=None, n_dist=10**6, verbose=False):
    """Calculate the average donor-acceptor distance <R_DA>
    between donor and acceptor

    Parameters
    ----------
    acv1, acv2 : fretraj.cloud.ACV
        donor and acceptor ACVs
    R_DA : ndarray
        ndarray of length n_dist containing distances between acv1 and acv2
    n_dist : int, default=10E6
        number of distances to calculate
    verbose: bool

    Returns
    -------
    float
        average distance between donor and acceptor
    """
    if R_DA is None:
        R_DA = dists_DA(acv1, acv2, n_dist)
        if verbose:
            print('> calculating R_DAs')
    if R_DA.ndim > 1:
        mean_R_DA = np.dot(R_DA[:, 0], R_DA[:, 1]) / R_DA[:, 1].sum()
    else:
        np.mean(R_DA)
    return mean_R_DA


def mean_dist_DA_ll(acv1, acv2, n_dist=10**6):
    """Calculate the average donor-acceptor distance <R_DA> between
    acv1 and acv2 using LabelLib

    Parameters
    ----------
    acv1, acv2 : fretraj.cloud.ACV
        donor and acceptor ACVs
    n_dist : int, default=10E6
        number of distances to calculate

    Returns
    -------
    float
        average distance between donor and acceptor
    """
    mean_R_DA = ll.meanDistance(acv1.ll_Grid3D, acv2.ll_Grid3D, n_dist)
    return mean_R_DA


def std_dist_DA(acv1, acv2, R_DA=None, n_dist=10**6, verbose=False):
    """Calculate the standard deviation of the donor acceptor distances sigma_R_DA

    Parameters
    ----------
    acv1, acv2 : fretraj.cloud.ACV
        donor and acceptor ACVs
    R_DA : ndarray
        ndarray of length n_dist containing distances between acv1 and acv2
    n_dist : int, default=10E6
        number of distances to calculate
    verbose : bool, default=False
        be verbose about calculation steps

    Returns
    -------
    float
        standard deviation of donor acceptor distances
    """
    if R_DA is None:
        R_DA = dists_DA(acv1, acv2, n_dist)
        if verbose:
            print('> calculating R_DAs')
    if R_DA.ndim > 1:
        sigma_R_DA = np.sqrt(np.dot(R_DA[:, 0]**2, R_DA[:, 1])
                             - np.dot(R_DA[:, 0], R_DA[:, 1])**2)
    else:
        sigma_R_DA = np.std(R_DA)
    return sigma_R_DA


def FRET_DA(acv1, acv2, R_DA=None, R0=54, n_dist=10**6, verbose=False):
    """Calculate the FRET efficiencies E for each donor acceptor distance

    Parameters
    ----------
    acv1, acv2 : fretraj.cloud.ACV
        donor and acceptor ACVs
    R_DA : ndarray, default=None
        ndarray of length n_dist containing distances between acv1 and acv2
        (if None calculate on the fly)
    R0 : float, default=54
        Förster radius in Angstrom
    n_dist : int, default=10E6
        number of distances to calculate
    verbose : bool, default=False
        be verbose about calculation steps

    Returns
    -------
    ndarray
        ndarray of shape 1 x n_dist or 2 x n_dist (if weights are included)
        containing FRET efficiencies between acv1 and acv2

    Notes
    -----
    The transfer efficiency *E* is given by

    .. math::  E = \\frac{R_0^6}{R_0^6+R_\\text{DA}^6}

    """
    if R_DA is None:
        R_DA = dists_DA(acv1, acv2, n_dist)
        if verbose:
            print('> calculating R_DAs')
    if R_DA.ndim > 1:
        E_DA = np.vstack((R0**6 / (R0**6 + R_DA[:, 0]**6), R_DA[:, 1])).T
    else:
        E_DA = R0**6 / (R0**6 + R_DA**6)
    return E_DA


def mean_FRET_DA(acv1, acv2, E_DA=None, R_DA=None, R0=54, n_dist=10**6,
                 verbose=False):
    """Calculate the average FRET efficiency <E> between donor and acceptor cloud

    Parameters
    ----------
    acv1, acv2 : fretraj.cloud.ACV
        donor and acceptor ACVs
    E_DA : ndarray, default=None
        ndarray of length n_dist containing FRET efficiencies
        between acv1 and acv2 (if None calculate on the fly)
    R_DA : ndarray, default=None
        ndarray of length n_dist containing distances
        between acv1 and acv2 (if None calculate on the fly)
    R0 : float, default=54
        Förster radius in Angstrom
    n_dist : int, default=10E6
        number of distances to calculate
    verbose : bool, default=False
        be verbose about calculation steps

    Returns
    -------
    float
        average FRET efficiency between donor and acceptor cloud
    """
    if E_DA is None:
        E_DA = FRET_DA(acv1, acv2, R_DA, R0, n_dist, verbose)
        if verbose:
            print('> calculating E_DAs')
    if E_DA.ndim > 1:
        mean_E_DA = np.dot(E_DA[:, 0], E_DA[:, 1]) / E_DA[:, 1].sum()
    else:
        mean_E_DA = np.mean(E_DA)
    return mean_E_DA


def std_FRET_DA(acv1, acv2, E_DA=None, R_DA=None, R0=54, n_dist=10**6,
                verbose=False):
    """Calculate the standard deviation of the FRET efficiencies sigma_E

    Parameters
    ----------
    acv1, acv2 : fretraj.cloud.ACV
        donor and acceptor ACVs
    E_DA : ndarray, default=None
        ndarray of length n_dist containing FRET efficiencies
        between acv1 and acv2 (if None calculate on the fly)
    R_DA : ndarray, default=None
        ndarray of length n_dist containing distances
        between acv1 and acv2 (if None calculate on the fly)
    R0 : float, default=54
        Förster radius in Angstrom
    n_dist : int, default=10E6
        number of distances to calculate
    verbose : bool, default=False
        be verbose about calculation steps

    Returns
    -------
    float
        standard deviation of the FRET efficiencies
    """
    if E_DA is None:
        E_DA = FRET_DA(acv1, acv2, R_DA, R0, n_dist, verbose)
        if verbose:
            print('> calculating E_DAs')
    if E_DA.ndim > 1:
        sigma_E_DA = np.sqrt(np.dot(E_DA[:, 0]**2, E_DA[:, 1])
                             - np.dot(E_DA[:, 0], E_DA[:, 1])**2)
    else:
        sigma_E_DA = np.std(E_DA)
    return sigma_E_DA


def mean_FRET_DA_ll(acv1, acv2, R0=54, n_dist=10**6):
    """Calculate the average FRET efficiency <E> between donor and acceptor cloud
    using LabelLib

    Parameters
    ----------
    acv1, acv2 : fretraj.cloud.ACV
        donor and acceptor ACVs
    R0 : float, default=54
        Förster radius in Angstrom
    n_dist : int, default=10E6
        number of distances to calculate

    Returns
    -------
    float
        average FRET efficiency between donor and acceptor cloud
    """
    return ll.meanEfficiency(acv1.ll_Grid3D, acv2.ll_Grid3D, R0, n_dist)


def mean_dist_DA_fromFRET(acv1, acv2, mean_E_DA=None, E_DA=None, R_DA=None,
                          R0=54, n_dist=10**6, verbose=False):
    """Calculate the FRET averaged donor acceptor distance <R_DA>_E

    Parameters
    ----------
    acv1, acv2 : fretraj.cloud.ACV
        donor and acceptor ACVs
    mean_E_DA : float, default=None
        average FRET efficiency between donor and acceptor cloud
        (if None calculate on the fly)
    E_DA : ndarray, default=None
        ndarray of length n_dist containing FRET efficiencies
        between acv1 and acv2 (if None calculate on the fly)
    R_DA : ndarray, default=None
        ndarray of length n_dist containing distances
        between acv1 and acv2 (if None calculate on the fly)
    R0 : float, default=54
        Förster radius in Angstrom
    n_dist : int, default=10E6
        number of distances to calculate
    verbose : bool, default=False
        be verbose about calculation steps

    Returns
    -------
    float
        FRET averaged donor acceptor distance

    Examples
    --------

    with known :math:`\\langle E_\\text{DA}\\rangle` (fastest)

    >>> mean_dist_DA_fromFRET(acv1, acv2, mean_E_DA=mymean_E_DA)

    with known :math:`E_\\text{DA}`, i.e. :math:`\\langle E_\\text{DA}\\rangle` is calculated on-the-fly

    >>> mean_dist_DA_fromFRET(acv1, acv2, E_DA=myE_DA)

    with known :math:`R_\\text{DA}`, i.e. :math:`E_\\text{DA}` and :math:`\\langle E_\\text{DA}\\rangle` are calculated on-the-fly

    >>> mean_dist_DA_fromFRET(acv1, acv2, R_DA=myR_DA)


    """
    if mean_E_DA is None:
        mean_E_DA = mean_FRET_DA(acv1, acv2, E_DA, R_DA, R0, n_dist, verbose)
        if verbose:
            print('> calculating mean_E_DA')
    mean_R_DA_E = R0 * (1 / mean_E_DA - 1)**(1 / 6)
    return mean_R_DA_E


def std_dist_DA_fromFRET(acv1, acv2, mean_E_DA=None, sigma_E_DA=None,
                         E_DA=None, R_DA=None, R0=54, n_dist=10**6,
                         verbose=False):
    """Calculate the standard deviation of the FRET averaged donor acceptor
    distances sigma_R_DA_E by error propagation

    Parameters
    ----------
    acv1, acv2 : fretraj.cloud.ACV
        donor and acceptor ACVs
    mean_E_DA : float, default=None
        average FRET efficiency between donor and acceptor cloud
        (if None calculate on the fly)
    sigma_E_DA : float, default=None
        standard deviation of the FRET efficiencies
    E_DA : ndarray, default=None
        ndarray of length n_dist containing FRET efficiencies
        between acv1 and acv2 (if None calculate on the fly)
    R_DA : ndarray, default=None
        ndarray of length n_dist containing distances
        between acv1 and acv2 (if None calculate on the fly)
    R0 : float, default=54
         Förster radius in Angstrom
    n_dist : int, default=10E6
        number of distances to calculate
    verbose : bool, default=False
        be verbose about calculation steps

    Returns
    -------
    float
        standard deviation of the FRET averaged donor acceptor distances
    """
    if (mean_E_DA is None) or (sigma_E_DA is None):
        mean_E_DA = mean_FRET_DA(acv1, acv2, E_DA, R_DA, R0, n_dist, verbose)
        sigma_E_DA = std_FRET_DA(acv1, acv2, E_DA, R_DA, R0, n_dist, verbose)
        if verbose:
            print('> calculating mean_E_DA and sigma_E_DA')
    # error propagation: sqrt( (df/dx)^2 * sigma_x^2 )
    sigma_R_DA_E = np.sqrt((R0/(6 * (1/mean_E_DA-1)**(5/6) * mean_E_DA**2))**2
                           * sigma_E_DA**2)
    return sigma_R_DA_E


def dist_attach(attach1_xyz, attach2_xyz):
    """Calculate the distance between the attachment sites

    Parameters
    ----------
    attach1_xyz, attach2_xyz : ndarray
        one-dimensional array of x-,y-,z-coordinates of the attachment points

    Returns
    -------
    float
        distance between the attachment sites
    """
    R_attach = np.sqrt(sum((attach1_xyz - attach2_xyz)**2))
    return R_attach


def R_DAE_to_Rmp(mean_R_DA_E):
    """Conversion function for <R_DA,E> to R_mp derived in Kalinin, *Nat. Methods* (2012)

    Parameters
    ----------
    mean_R_DA_E : float
        FRET averaged donor acceptor distance

    Returns
    -------
    float
        distance between the mean dye positions

    Notes
    -----
    The conversion function between :math:`\\langle R_\\text{DA}\\rangle_E` and :math:`R_\\text{mp}` is a third-order 
    polynomial:

    .. math::  R_\\text{mp} = 1.109\\cdot 10^{-5}\\cdot\\langle R_\\text{DA}\\rangle_E^3 - 7.286\\cdot 10^{-3}\\cdot
        \\langle R_\\text{DA}\\rangle_E^2 + 1.979\\cdot\\langle R_\\text{DA}\\rangle_E - 34.345

    References
    ----------

    .. [3] Kalinin, S. et al. "A toolkit and benchmark study for FRET-restrained high-precision \
        structural modeling", *Nat. Methods* **9**, 1218–1225 (2012).
    """
    R_mp = 1.109*10**-5*mean_R_DA_E**3 - 7.286*10**-3*mean_R_DA_E**2
    + 1.979*mean_R_DA_E - 34.345
    return R_mp



#!/usr/bin/env python3
import pytest
import numpy as np
from fretraj import export
from fretraj import cloud


CLOUD_XYZQT = np.array([[0.5, 1.0, 0.5, 1.0, 1],
                        [0.5, 1.0, 0.5, 8.0, 2]])
MP = CLOUD_XYZQT[0, 0:3]
ELEMENT = export.ELEMENT
ELEMENT_MP = export.ELEMENT_MP
NAME = export.NAME
NAME_MP = export.ELEMENT_MP


def test_xyz():
    s = export.xyz(CLOUD_XYZQT, MP)
    assert s.split('\n')[2] == f'{ELEMENT}\t0.500\t1.000\t0.500\t1.000'
    assert s.split('\n')[4] == f'{ELEMENT_MP}\t0.500\t1.000\t0.500\t-1.000'


def test_xyz_encodeElem():
    s = export.xyz(CLOUD_XYZQT, MP, encode_element=True)
    assert s.split('\n')[2] == 'F\t0.500\t1.000\t0.500\t1.000'
    assert s.split('\n')[3] == 'C\t0.500\t1.000\t0.500\t8.000'


def test_xyz_noweights():
    s = export.xyz(CLOUD_XYZQT, MP, write_weights=False)
    assert s.split('\n')[2] == f'{ELEMENT}\t0.500\t1.000\t0.500'
    assert s.split('\n')[4] == f'{ELEMENT_MP}\t0.500\t1.000\t0.500'


def test_xyz_noweights_encodeElem():
    s = export.xyz(CLOUD_XYZQT, MP, encode_element=True, write_weights=False)
    assert s.split('\n')[2] == 'F\t0.500\t1.000\t0.500'
    assert s.split('\n')[3] == 'C\t0.500\t1.000\t0.500'


def test_pdb():
    s = export.pdb(CLOUD_XYZQT, MP)
    assert s.split('\n')[0] == f'ATOM      1    {NAME}  FV     1       0.500   1.000   0.500 99.00  1.00           {ELEMENT}  '
    assert s.split('\n')[2] == f'ATOM      3   {NAME_MP}  MP     0       0.500   1.000   0.500 99.00 -1.00          {ELEMENT_MP}  '


def test_openDX():
    grid_3d = np.zeros((3, 3, 3))
    grid_3d[(0, 0, 0)] = 1
    grid_3d[(0, 0, 1)] = 10
    s = export.open_dx(grid_3d, (0, 0, 0), (1, 1, 1))
    assert s.split('\n')[1] == 'origin 0.000 0.000 0.000'
    assert s.split('\n')[7] == '1.0 10.0 0.0'


if __name__ == '__main__':
    pytest.main()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['fretraj']

package_data = \
{'': ['*'], 'fretraj': ['UI/*', 'examples/*', 'skripts/*']}

install_requires = \
['PyQt5>=5.12.3,<6.0.0',
 'astunparse==1.6.2',
 'importlib-metadata>=3.7.2,<4.0.0',
 'mdtraj>=1.9.5,<2.0.0',
 'nglview>=2.7.7,<3.0.0',
 'numba<=0.50.1',
 'numpy<1.20.0',
 'pandas>=1.2.2,<2.0.0']

entry_points = \
{'console_scripts': ['fretraj = fretraj.cloud:main',
                     'fretraj_gui = fretraj.fretraj_gui:main',
                     'pymol_vis = fretraj.console:pymol_vis',
                     'vmd_vis = fretraj.console:vmd_vis']}

setup_kwargs = {
    'name': 'fretraj',
    'version': '0.1.5',
    'description': 'Predicting FRET with accessible-contact volumes',
    'long_description': '<img src="https://raw.githubusercontent.com/fdsteffen/fretraj/master/docs/images/fretraj_logo_readme.png">\n\n[![Build Status](https://github.com/fdsteffen/fretraj/workflows/FRETraj%20build/badge.svg)](https://github.com/fdsteffen/fretraj/actions)\n[![Docs Status](https://github.com/fdsteffen/fretraj/workflows/FRETraj%20docs/badge.svg)](https://github.com/fdsteffen/fretraj/actions)\n[![PyPI](https://img.shields.io/pypi/v/fretraj)](https://pypi.org/project/fretraj/)\n[![Anaconda-Server Badge](https://img.shields.io/badge/Install%20with-conda-green.svg)](https://anaconda.org/rna-fretools/fretraj)\n[![codecov](https://codecov.io/gh/fdsteffen/fretraj/branch/master/graph/badge.svg?token=A2E70FbycQ)](https://codecov.io/gh/fdsteffen/fretraj)\n\n*FRETraj* is a Python module for calculating **multiple accessible-contact volumes** (multi-ACV) and predicting **FRET efficiencies**. It provides an interface to the [*LabelLib*](https://github.com/Fluorescence-Tools/LabelLib) library to simulate fluorophore dynamics. The package features a user-friendly **PyMOL plugin**<sup>[1](#pymol)</sup> which can be used to explore different labeling positions when designing FRET experiments. In an AV simulation the fluorophore distribution is estimated by a shortest path search (Djikstra algorithm) using a coarse-grained dye probe. *FRETraj* further implements a **Python-only** version of the geometrical clash search used in *LabelLib*. This should facilitate prototyping of new features for the ACV algorithm.\n\n<img src="https://raw.githubusercontent.com/fdsteffen/fretraj/master/docs/images/graphical_abstract.png">\n     \nA recent addition to the original AV model (Kalinin et al. *Nat. Methods*, 2012) is the so-called **contact volume** (Steffen et. al. *PCCP* 2016). Here, the accessible volume is split into a free volume (FV, transparent) where the dye is freely diffusing and a contact volume (CV, opaque) where the dye stacks to the biomolecular surface. Time-resolved anisotropy experiments suggest that certain fluorophores, among those the commonly used cyanines Cy3 and Cy5, are particularly prone to interact with both proteins and nucleic acids. The contact volume accounts for this effect by reweighting the point-cloud. By choosing different experimental weights for the free and contact component the AV dye model is refined, making *in silico* FRET predictions more reliable.\n\n## Installation\nFollow the instructions for your platform [here](https://rna-fretools.github.io/fretraj/getting_started/installation)\n\n## References\nIf you use **FRETraj** in your work please refer to the following paper:\n\n- F.D. Steffen, R.K.O. Sigel, R. Börner, *Phys. Chem. Chem. Phys.* **2016**, *18*, 29045-29055. [![](https://img.shields.io/badge/DOI-10.1039/C6CP04277E-blue.svg)](https://doi.org/10.1039/C6CP04277E)\n\n### Additional readings\n- S. Kalinin, T. Peulen, C.A.M. Seidel et al. *Nat. Methods*, **2012**, *9*, 1218-1225.\n- T. Eilert, M. Beckers, F. Drechsler, J. Michaelis, *Comput. Phys. Commun.*, **2017**, *219*, 377–389.\n- M. Dimura, T. Peulen, C.A.M. Seidel et al. *Curr. Opin. Struct. Biol.* **2016**, *40*, 163-185.\n- M. Dimura, T. Peulen, C.A.M Seidel et al. *Nat. Commun.* **2020**, *11*, 5394.\n\n---\n\n<sup><a name="pymol">1</a></sup> PyMOL is a trademark of Schrödinger, LLC.\n',
    'author': 'Fabio Steffen',
    'author_email': 'fabio.steffen@chem.uzh.ch',
    'maintainer': 'Fabio Steffen',
    'maintainer_email': 'fabio.steffen@chem.uzh.ch',
    'url': 'https://rna-fretools.github.io/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<3.9',
}


setup(**setup_kwargs)



try:
    import importlib.metadata as meta  # Python >=3.8
except ModuleNotFoundError:
    import importlib_metadata as meta  # Python 3.7

try:
    metadata = meta.metadata('fretraj')
except meta.PackageNotFoundError:
    print('FRETraj is not installed yet.')

__version__ = metadata['Version']
__author__ = metadata['Author']
__keywords__ = metadata['Keywords']
__license__ = metadata['License']

from . import cloud
from . import export
from . import fret
from . import grid
from . import jupyter

import warnings
import numba as nb

warnings.simplefilter('ignore', category=nb.errors.NumbaDeprecationWarning)
warnings.simplefilter('ignore', category=nb.errors.NumbaPendingDeprecationWarning)



#!/usr/bin/env python3
import pytest
import os
import argparse
import json
from fretraj import cloud


MINIMAL_LABELS = {"Position":
                  {"Cy3":
                   {"attach_id": 1,
                    "linker_length": 20,
                    "linker_width": 5,
                    "dye_radius1": 8,
                    "dye_radius2": 3,
                    "dye_radius3": 3,
                    },
                   "Cy5":
                       {"attach_id": 288,
                        "linker_length": 20,
                        "linker_width": 5,
                        "dye_radius1": 9.5,
                        "dye_radius2": 3,
                        "dye_radius3": 3,
                        },
                   },
                  "Distance": {"Cy3-Cy5":
                               {"R0": 54}
                               }
                  }


@pytest.fixture
def complete_labels():
    cloud.check_labels(MINIMAL_LABELS, verbose=False)
    return MINIMAL_LABELS


def test_simulation_type(complete_labels):
    assert complete_labels['Position']['Cy3']['simulation_type'] == 'AV3'


@pytest.mark.parametrize('label_dict, missing_item, error',
                         [(MINIMAL_LABELS, 'Position', ValueError),
                          (MINIMAL_LABELS['Position']['Cy3'], 'linker_length', KeyError)])
def test_missing_item(label_dict, missing_item, error, monkeypatch):
    monkeypatch.delitem(label_dict, missing_item)
    with pytest.raises(error):
        cloud.check_labels(MINIMAL_LABELS, verbose=False)


def test_attachID_wrongtype(monkeypatch):
    monkeypatch.setitem(MINIMAL_LABELS['Position']['Cy3'], 'linker_length', '1')
    with pytest.raises(TypeError):
        cloud.check_labels(MINIMAL_LABELS, verbose=False)


def test_unrecognized_key(monkeypatch):
    monkeypatch.setitem(MINIMAL_LABELS['Position']['Cy3'], 'unknown_key', 'some_value')
    with pytest.raises(KeyError):
        cloud.check_labels(MINIMAL_LABELS, verbose=False)


def test_AV1(complete_labels, monkeypatch):
    monkeypatch.setitem(MINIMAL_LABELS['Position']['Cy3'], 'simulation_type', 'AV1')
    cloud.check_labels(MINIMAL_LABELS, verbose=False)
    assert complete_labels['Position']['Cy3']['dye_radius2'] == 0


def test_parseCmd(mocker):
    mocker.patch('argparse.ArgumentParser.parse_args',
                 return_value=argparse.Namespace(input='test.pdb', parameters='param.json', output=None))
    out = cloud.parseCmd()
    assert out == ('test.pdb', 'param.json', None)


def test_labelingParams_valid(mocker):
    mocker.patch('fretraj.cloud.open')
    mocker.patch('json.load', return_value=MINIMAL_LABELS)
    labels_json = cloud.labeling_params('param.json')
    assert labels_json['Position']['Cy3']['attach_id'] == 1


def test_labelingParams_missing_key(mocker, monkeypatch):
    mocker.patch('fretraj.cloud.open')
    monkeypatch.delitem(MINIMAL_LABELS['Position']['Cy3'], 'linker_length')
    mocker.patch('json.load', return_value=MINIMAL_LABELS)
    labels_json = cloud.labeling_params('param.json')
    assert labels_json is None


def test_labelingParams_wrongtype(mocker, monkeypatch):
    mocker.patch('fretraj.cloud.open')
    monkeypatch.setitem(MINIMAL_LABELS['Position']['Cy3'], 'linker_length', '1')
    mocker.patch('json.load', return_value=MINIMAL_LABELS)
    labels_json = cloud.labeling_params('param.json')
    assert labels_json is None


if __name__ == '__main__':
    pytest.main()
